#!/usr/bin/env python
"""
Update the Unicode code tables for wcwidth.

This is code generation using jinja2. This is typically executed through tox,

$ tox -e update

https://github.com/jquast/wcwidth
"""

from __future__ import annotations

# std imports
import io
import os
import re
import sys
import glob
import string
import hashlib
import zipfile
import argparse
import datetime
import textwrap
import functools
import unicodedata
from pathlib import Path
from dataclasses import field, fields, dataclass

from typing import Any, Mapping, Iterable, Iterator, Optional, Sequence, Collection

try:
    from typing import Self
except ImportError:
    from typing_extensions import Self

# 3rd party
import yaml
import jinja2
import requests
import urllib3.util
import dateutil.parser

try:
    # 3rd party
    from yaml import CSafeLoader as SafeLoader
except ImportError:
    # 3rd party
    from yaml import SafeLoader


EXCLUDE_VERSIONS = ['2.0.0', '2.1.2', '3.0.0', '3.1.0', '3.2.0', '4.0.0']

PATH_UP = os.path.relpath(os.path.join(os.path.dirname(__file__), os.path.pardir))
PATH_DATA = os.path.join(PATH_UP, 'data')
PATH_TESTS = os.path.join(PATH_UP, 'tests')
PATH_UCS_DETECT_DATA = os.path.join(PATH_UP, 'ucs-detect', 'data')
# "wcwidth/bin/update-tables.py", even on Windows
# not really a path, if the git repo isn't named "wcwidth"
THIS_FILEPATH = ('wcwidth/' +
                 Path(__file__).resolve().relative_to(Path(PATH_UP).resolve()).as_posix())

JINJA_ENV = jinja2.Environment(
    loader=jinja2.FileSystemLoader(os.path.join(PATH_UP, 'code_templates')),
    keep_trailing_newline=True)
UTC_NOW = datetime.datetime.now(datetime.UTC).strftime("%Y-%m-%d %H:%M:%S UTC")

CONNECT_TIMEOUT = int(os.environ.get('CONNECT_TIMEOUT', '10'))
READ_TIMEOUT = int(os.environ.get('READ_TIMEOUT', '30'))
FETCH_BLOCKSIZE = int(os.environ.get('FETCH_BLOCKSIZE', '4096'))
MAX_RETRIES = int(os.environ.get('MAX_RETRIES', '10'))
BACKOFF_FACTOR = float(os.environ.get('BACKOFF_FACTOR', '1.0'))

# Global flag set by main() from --check-last-modified CLI argument.
# When True, perform HTTP HEAD requests to check if remote files are newer.
# Default is False because Unicode data files rarely change once published.
CHECK_LAST_MODIFIED = False

# Hangul Jamo is a decomposed form of Hangul Syllables, see
# see https://www.unicode.org/faq/korean.html#3
#     https://github.com/ridiculousfish/widecharwidth/pull/17
#     https://github.com/jquast/ucs-detect/issues/9
#     https://devblogs.microsoft.com/oldnewthing/20201009-00/?p=104351
# "Conjoining Jamo are divided into three classes: L, V, T (Leading
#  consonant, Vowel, Trailing consonant). A Hangul Syllable consists of
#  <LV> or <LVT> sequences."
HANGUL_JAMO_ZEROWIDTH = (
    *range(0x1160, 0x1200),  # Hangul Jungseong Filler .. Hangul Jongseong Ssangnieun
    *range(0xD7B0, 0xD800),  # Hangul Jungseong O-Yeo  .. Undefined Character of Hangul Jamo Extended-B
)

# Variation Selector-15 and 16
HEX_STR_VS15, HEX_STR_VS16 = ('FE0E', 'FE0F')

# Grapheme Break Property values from UAX #29
GRAPHEME_BREAK_PROPERTIES = (
    'CR', 'LF', 'Control', 'Extend', 'ZWJ', 'Regional_Indicator',
    'Prepend', 'SpacingMark', 'L', 'V', 'T', 'LV', 'LVT'
)
INCB_VALUES = ('Linker', 'Consonant', 'Extend')


def _bisearch(ucs, table):
    """A copy of wcwwidth._bisearch."""
    # to prevent having issues when depending on code that imports our generated code.
    lbound = 0
    ubound = len(table) - 1
    if ucs < table[0][0] or ucs > table[ubound][1]:
        return 0
    while ubound >= lbound:
        mid = (lbound + ubound) // 2
        if ucs > table[mid][1]:
            lbound = mid + 1
        elif ucs < table[mid][0]:
            ubound = mid - 1
        else:
            return 1
    return 0


@dataclass(order=True, frozen=True)
class UnicodeVersion:
    """A class for comparable unicode version."""

    major: int
    minor: int
    micro: Optional[int]

    @classmethod
    def parse(cls, version_str: str) -> UnicodeVersion:
        """
        Parse a version string.

        >>> UnicodeVersion.parse("14.0.0")
        UnicodeVersion(major=14, minor=0, micro=0)
        """
        ver_ints = tuple(map(int, version_str.split(".")[:3]))
        return cls(major=ver_ints[0], minor=ver_ints[1],
                   micro=ver_ints[2] if len(ver_ints) > 2 else None)

    def __str__(self) -> str:
        """
        >>> str(UnicodeVersion(12, 1, 0))
        '12.1.0'
        """
        maybe_micro = ''
        if self.micro is not None:
            maybe_micro = f'.{self.micro}'
        return f'{self.major}.{self.minor}{maybe_micro}'


@dataclass(frozen=True)
class TableEntry:
    """An entry of a unicode table."""

    code_range: Optional[tuple[int, int]]
    properties: tuple[str, ...]
    comment: str

    def filter_by_category_width(self, wide: int) -> bool:
        """
        Return whether entry matches displayed width.

        Parses both DerivedGeneralCategory.txt and EastAsianWidth.txt
        """
        if self.code_range is None:
            return False
        elif self.properties[0] == 'Sk':
            if 'EMOJI MODIFIER' in self.comment:
                # Standalone Fitzpatrick modifiers display as wide (2 cells).
                # Zero-width when following an emoji base is handled contextually
                # in wcswidth() and width().
                return wide == 2
            elif 'FULLWIDTH' in self.comment:
                # Some codepoints in 'Sk' categories are fullwidth(!)
                # at this time just 3, FULLWIDTH: CIRCUMFLEX ACCENT, GRAVE ACCENT, and MACRON
                return wide == 2
            else:
                # the rest are narrow
                return wide == 1
        # Me Enclosing Mark
        # Mn Nonspacing Mark
        # Cf Format
        # Zl Line Separator
        # Zp Paragraph Separator
        if self.properties[0] in ('Me', 'Mn', 'Mc', 'Cf', 'Zl', 'Zp'):
            return wide == 0
        # F  Fullwidth
        # W  Wide
        if self.properties[0] in ('W', 'F'):
            return wide == 2
        return wide == 1

    @staticmethod
    def parse_width_category_values(table_iter: Iterator[TableEntry],
                                    wide: int) -> set[tuple[int, int]]:
        """Parse value ranges of unicode data files, by given category and width."""
        return {n
                for entry in table_iter
                if entry.filter_by_category_width(wide)
                for n in list(range(entry.code_range[0], entry.code_range[1]))}


@dataclass
class TableDef:
    filename: str
    date: str
    values: set[int]

    def as_value_ranges(self) -> list[tuple[int, int]]:
        """Return a list of tuple of (start, end) ranges for given set of 'values'."""
        table: list[tuple[int, int]] = []
        values_iter = iter(sorted(self.values))
        start = end = next(values_iter)
        table.append((start, end))

        for value in values_iter:
            # remove last-most entry for comparison,
            start, end = table.pop()
            if end == value - 1:
                # continuation of existing range, rewrite
                table.append((start, value,))
            else:
                # non-continuation: insert back previous range,
                table.append((start, end,))
                # and start a new one
                table.append((value, value,))
        return table

    @property
    def hex_range_descriptions(self) -> list[tuple[str, str, str]]:
        """Convert integers into string table of (hex_start, hex_end, txt_description)."""
        pytable_values: list[tuple[str, str, str]] = []
        for start, end in self.as_value_ranges():
            hex_start, hex_end = f'0x{start:05x}', f'0x{end:05x}'
            ucs_start, ucs_end = chr(start), chr(end)
            name_start = name_ucs(ucs_start) or '(nil)'
            name_end = name_ucs(ucs_end) or '(nil)'
            if name_start != name_end:
                txt_description = f'{name_start[:24].rstrip():24s}..{name_end[:24].rstrip()}'
            else:
                txt_description = f'{name_start[:48]}'
            pytable_values.append((hex_start, hex_end, txt_description))
        return pytable_values


@dataclass(frozen=True)
class RenderContext:
    def to_dict(self) -> dict[str, Any]:
        return {field.name: getattr(self, field.name)
                for field in fields(self)}


@dataclass(frozen=True)
class UnicodeVersionPyRenderCtx(RenderContext):
    versions: Collection[UnicodeVersion]


@dataclass(frozen=True)
class UnicodeVersionRstRenderCtx(RenderContext):
    source_headers: Sequence[tuple[str, str]]


@dataclass(frozen=True)
class UnicodeTableRenderCtx(RenderContext):
    variable_name: str
    table: Mapping[UnicodeVersion, TableDef]


@dataclass
class RenderDefinition:
    """Base class, do not instantiate it directly."""

    jinja_filename: str
    output_filename: str
    render_context: RenderContext

    _template: jinja2.Template = field(init=False, repr=False)
    _render_context: dict[str, Any] = field(init=False, repr=False)

    def __post_init__(self) -> None:
        self._template = JINJA_ENV.get_template(self.jinja_filename)
        self._render_context = {
            'utc_now': UTC_NOW,
            'this_filepath': THIS_FILEPATH,
            **self.render_context.to_dict(),
        }

    def render(self) -> str:
        """Just like jinja2.Template.render."""
        return self._template.render(self._render_context)

    def generate(self) -> Iterator[str]:
        """Just like jinja2.Template.generate."""
        return self._template.generate(self._render_context)


@dataclass
class UnicodeVersionPyRenderDef(RenderDefinition):
    render_context: UnicodeVersionPyRenderCtx

    @classmethod
    def new(cls, context: UnicodeVersionPyRenderCtx) -> Self:
        return cls(
            jinja_filename='unicode_versions.py.j2',
            output_filename=os.path.join(PATH_UP, 'wcwidth', 'unicode_versions.py'),
            render_context=context,
        )


@dataclass
class UnicodeVersionRstRenderDef(RenderDefinition):
    render_context: UnicodeVersionRstRenderCtx

    @classmethod
    def new(cls, context: UnicodeVersionRstRenderCtx) -> Self:
        return cls(
            jinja_filename='unicode_version.rst.j2',
            output_filename=os.path.join(PATH_UP, 'docs', 'unicode_version.rst'),
            render_context=context,
        )


@dataclass
class UnicodeTableRenderDef(RenderDefinition):
    render_context: UnicodeTableRenderCtx

    @classmethod
    def new(cls, filename: str, context: UnicodeTableRenderCtx) -> Self:
        _, ext = os.path.splitext(filename)
        if ext == '.py':
            jinja_filename = 'python_table.py.j2'
        elif ext == '.c':
            # TODO
            jinja_filename = 'c_table.c.j2'
        else:
            raise ValueError('filename must be a Python or a C file')

        return cls(
            jinja_filename=jinja_filename,
            output_filename=os.path.join(PATH_UP, 'wcwidth', filename),
            render_context=context,
        )


@dataclass(frozen=True)
class GraphemeTableRenderCtx(RenderContext):
    """Render context for grapheme tables (latest version only)."""

    unicode_version: str
    tables: Mapping[str, TableDef]


@dataclass
class GraphemeTableRenderDef(RenderDefinition):
    render_context: GraphemeTableRenderCtx

    @classmethod
    def new(cls, context: GraphemeTableRenderCtx) -> Self:
        return cls(
            jinja_filename='grapheme_table.py.j2',
            output_filename=os.path.join(PATH_UP, 'wcwidth', 'table_grapheme.py'),
            render_context=context,
        )


@functools.cache
def fetch_unicode_versions() -> list[UnicodeVersion]:
    """Fetch, determine, and return Unicode Versions for processing."""
    pattern = re.compile(r'#.*assigned in Unicode ([0-9.]+)')
    versions: list[UnicodeVersion] = []
    with open(UnicodeDataFile.DerivedAge(), encoding='utf-8') as f:
        for line in f:
            if match := re.match(pattern, line):
                version = match.group(1)
                if version not in EXCLUDE_VERSIONS:
                    versions.append(UnicodeVersion.parse(version))
    versions.sort()
    return versions


def fetch_source_headers() -> UnicodeVersionRstRenderCtx:
    headers: list[tuple[str, str]] = []
    for filename in UnicodeDataFile.filenames():
        header_description = cite_source_description(filename)
        headers.append(header_description)
    return UnicodeVersionRstRenderCtx(headers)


def fetch_table_wide_data() -> UnicodeTableRenderCtx:
    """Fetch east-asian tables for the latest Unicode version only."""
    table: dict[UnicodeVersion, TableDef] = {}
    version = fetch_unicode_versions()[-1]  # Only latest version

    # parse typical 'wide' characters by categories 'W' and 'F',
    table[version] = parse_category(fname=UnicodeDataFile.EastAsianWidth(version),
                                    wide=2)

    # subtract(!) wide characters that were defined above as 'W' category in EastAsianWidth,
    # but also zero-width category 'Mn' or 'Mc' in DerivedGeneralCategory!
    table[version].values = table[version].values.difference(parse_category(
        fname=UnicodeDataFile.DerivedGeneralCategory(version),
        wide=0).values)

    # Also subtract Hangul Jamo Vowels and Hangul Trailing Consonants
    table[version].values = table[version].values.difference(HANGUL_JAMO_ZEROWIDTH)

    # Subtract Default_Ignorable_Code_Point characters (they should be zero-width).
    # Exception: U+115F HANGUL CHOSEONG FILLER remains wide for jamo composition.
    # See https://github.com/jquast/wcwidth/issues/118
    default_ignorable = set(parse_derived_core_property(
        fname=UnicodeDataFile.DerivedCoreProperties(version),
        property_name='Default_Ignorable_Code_Point'))
    default_ignorable.discard(0x115F)  # Keep HANGUL CHOSEONG FILLER as wide
    table[version].values = table[version].values.difference(default_ignorable)

    # finally, join with atypical 'wide' characters defined by category 'Sk',
    fname = UnicodeDataFile.DerivedGeneralCategory(version)
    table[version].values.update(parse_category(fname=fname, wide=2).values)

    # Add Regional Indicator symbols (U+1F1E6..U+1F1FF). Though classified as
    # Neutral in EastAsianWidth.txt, terminals universally render these as
    # double-width. Pairing (flag emoji) is handled contextually in wcswidth()
    # and width().
    table[version].values.update(range(0x1F1E6, 0x1F1FF + 1))

    return UnicodeTableRenderCtx('WIDE_EASTASIAN', table)


def fetch_table_zero_data() -> UnicodeTableRenderCtx:
    """
    Fetch zero width tables for the latest Unicode version only.

    See also: https://unicode.org/L2/L2002/02368-default-ignorable.html
    """
    table: dict[UnicodeVersion, TableDef] = {}
    version = fetch_unicode_versions()[-1]  # Only latest version

    # Determine values of zero-width character lookup table by the following category codes
    fname = UnicodeDataFile.DerivedGeneralCategory(version)
    table[version] = parse_category(fname=fname, wide=0)

    # Include NULL
    table[version].values.add(0)

    # Add Hangul Jamo Vowels and Hangul Trailing Consonants
    table[version].values.update(HANGUL_JAMO_ZEROWIDTH)

    # Add Default_Ignorable_Code_Point characters
    # Per Unicode Standard (https://www.unicode.org/faq/unsup_char.html):
    # "All default-ignorable characters should be rendered as completely invisible
    # (and non advancing, i.e. 'zero width'), if not explicitly supported in rendering."
    #
    # See also:
    # - https://www.unicode.org/reports/tr44/#Default_Ignorable_Code_Point
    # - https://github.com/jquast/wcwidth/issues/118
    table[version].values.update(parse_derived_core_property(
        fname=UnicodeDataFile.DerivedCoreProperties(version),
        property_name='Default_Ignorable_Code_Point'))

    # Remove U+115F HANGUL CHOSEONG FILLER from zero-width table.
    # Although it has Default_Ignorable_Code_Point property, it should remain
    # width 2 because it combines with other Hangul Jamo to form width-2
    # syllable blocks.
    table[version].values.discard(0x115F)

    # Remove u+00AD categoryCode=Cf name="SOFT HYPHEN",
    # > https://www.unicode.org/faq/casemap_charprop.html
    #
    # > Q: Unicode now treats the SOFT HYPHEN as format control (Cf)
    # > character when formerly it was a punctuation character (Pd).
    # > Doesn't this break ISO 8859-1 compatibility?
    #
    # > [..] In a terminal emulation environment, particularly in
    # > ISO-8859-1 contexts, one could display the SOFT HYPHEN as a hyphen
    # > in all circumstances.
    #
    # This value was wrongly measured as a width of '0' in this wcwidth
    # versions 0.2.9 - 0.2.13. Fixed in 0.2.14
    table[version].values.discard(0x00AD)  # SOFT HYPHEN

    # Remove Prepended_Concatenation_Mark characters from zero-width.
    # Per Unicode Standard Annex #44, these format characters (General_Category=Cf) have
    # mandatory visible display and should NOT be treated as invisible.
    # See https://github.com/jquast/wcwidth/issues/119
    table[version].values = table[version].values.difference(
        parse_derived_core_property(
            fname=UnicodeDataFile.PropList(version),
            property_name='Prepended_Concatenation_Mark'))

    # Remove Emoji Modifier Fitzpatrick types (U+1F3FB..U+1F3FF) from zero-width.
    # Standalone they display as wide (2 cells); they are only zero-width when
    # following an emoji base character in sequence, handled contextually in
    # wcswidth() and width().
    table[version].values -= set(range(0x1F3FB, 0x1F3FF + 1))

    return UnicodeTableRenderCtx('ZERO_WIDTH', table)


def fetch_table_category_mc_data() -> UnicodeTableRenderCtx:
    """
    Fetch Spacing Combining Mark (Mc) character table for the latest Unicode version.

    Characters with General_Category=Mc are combining marks that typically occupy a cell width when
    following a base character, but should be zero-width when standalone. This table is used for
    context-aware width measurement.
    """
    table: dict[UnicodeVersion, TableDef] = {}
    version = fetch_unicode_versions()[-1]

    fname = UnicodeDataFile.DerivedGeneralCategory(version)
    print(f'parsing {fname}, category=Mc: ', end='', flush=True)

    with open(fname, encoding='utf-8') as f:
        table_iter = parse_unicode_table(f)
        file_version = next(table_iter).comment.strip()
        date = next(table_iter).comment.split(':', 1)[1].strip()
        values = {n
                  for entry in table_iter
                  if entry.code_range is not None and entry.properties[0] == 'Mc'
                  for n in range(entry.code_range[0], entry.code_range[1])}
    print('ok')
    table[version] = TableDef(file_version, date, values)
    return UnicodeTableRenderCtx('CATEGORY_MC', table)


def fetch_table_ambiguous_data() -> UnicodeTableRenderCtx:
    """
    Fetch east-asian ambiguous character table for the latest Unicode version.

    East Asian Ambiguous (A) characters can display as either 1 cell (narrow) or 2 cells (wide)
    depending on the terminal's configuration. This table allows users to opt-in to treating these
    characters as wide by passing ambiguous_width=2 to wcwidth/wcswidth.
    """
    table: dict[UnicodeVersion, TableDef] = {}
    version = fetch_unicode_versions()[-1]
    # parse 'ambiguous' characters by category 'A'
    table[version] = parse_category_ambiguous(
        fname=UnicodeDataFile.EastAsianWidth(version)
    )
    # Subtract zero-width characters (they should remain zero-width
    # regardless of ambiguous_width setting)
    table[version].values = table[version].values.difference(
        parse_category(
            fname=UnicodeDataFile.DerivedGeneralCategory(version),
            wide=0
        ).values
    )
    return UnicodeTableRenderCtx('AMBIGUOUS_EASTASIAN', table)


def fetch_table_vs16_data() -> UnicodeTableRenderCtx:
    """
    Fetch and create a "narrow to wide variation-16" lookup table.

    Characters in this table are all narrow, but when combined with a variation
    selector-16 (\uFE0F), they become wide, for the given versions of unicode.

    UNICODE_VERSION=9.0.0 or greater is required to enable detection of the effect
    of *any* 'variation selector-16' narrow emoji becoming wide. Just two total
    files are parsed to create ONE unicode version table supporting all
    Unicode versions 9.0.0 and later.

    Because of the ambiguity of versions in these early emoji data files, which
    match unicode releases 8, 9, and 10, these specifications were mostly
    implemented only in Terminals supporting Unicode 9.0 or later.

    For that reason, and that **these values are not expected to change**,
    If they do, a noticeable change would occur in `wcwidth/table_vs16.py`
    falsely labeled under version 9.0 but is prevented by assertion.

    One example, where v3.2 became v1.1 ("-" 12.0, "+" 15.1)::

         -2620 FE0F  ; Basic_Emoji  ; skull and crossbones        #  3.2  [1] (☠️)
         +2620 FE0F  ; emoji style; # (1.1) SKULL AND CROSSBONES

    Or another discrepancy, published in unicode 12.0 as emoji version 5.2, but
    missing entirely in the emoji-variation-sequences.txt published with unicode
    version 15.1::

        26F3 FE0E  ; text style;  # (5.2) FLAG IN HOLE

    while some terminals display \\u0036\\uFE0F as a wide number one (kitty),
    others display as ascii 1 with a no-effect zero-width (iTerm2) and others
    have a strange narrow font corruption, I think it is fair to call these
    ambiguous, no doubt in part because of these issues, see related
    'ucs-detect' project.

    Note that version 3.2 became 1.1, which would change unicode release of 9.0
    to version 8.0.
    """
    table: dict[UnicodeVersion, TableDef] = {}
    unicode_latest = fetch_unicode_versions()[-1]

    wide_tables = fetch_table_wide_data().table
    unicode_version = UnicodeVersion.parse('9.0.0')

    # parse table formatted by the latest emoji release (developed with
    # 15.1.0) and parse a single file for all individual releases
    table[unicode_version] = parse_vs_data(fname=UnicodeDataFile.EmojiVariationSequences(unicode_latest),
                                           ubound_unicode_version=unicode_version,
                                           hex_str_vs=HEX_STR_VS16)

    # parse and join the final emoji release 12.0 of the earlier "type"
    table[unicode_version].values.update(
        parse_vs_data(fname=UnicodeDataFile.LegacyEmojiVariationSequences(),
                      ubound_unicode_version=unicode_version,
                      hex_str_vs=HEX_STR_VS16).values)

    # perform culling on any values that are already understood as 'wide'
    # without the variation-16 selector (use latest wide table)
    wide_table = wide_tables[unicode_latest].as_value_ranges()
    table[unicode_version].values = {
        ucs for ucs in table[unicode_version].values
        if not _bisearch(ucs, wide_table)
    }

    return UnicodeTableRenderCtx('VS16_NARROW_TO_WIDE', table)


def fetch_table_vs15_data() -> UnicodeTableRenderCtx:
    """
    Fetch and create a "wide to narrow variation-15" lookup table.

    Characters in this table are normally wide, but when combined with a variation
    selector-15 (\\uFE0E), they become narrow. The table is built from ''text style''
    entries in ``emoji-variation-sequences.txt``, filtered to only wide characters.

    This mirrors ``fetch_table_vs16_data``, which builds the ''narrow to wide''
    variation-16 table from ''emoji style'' entries.
    """
    table: dict[UnicodeVersion, TableDef] = {}
    unicode_latest = fetch_unicode_versions()[-1]

    wide_tables = fetch_table_wide_data().table
    unicode_version = UnicodeVersion.parse('9.0.0')

    # Parse FE0E (text style) entries from the latest emoji release
    table[unicode_version] = parse_vs_data(
        fname=UnicodeDataFile.EmojiVariationSequences(unicode_latest),
        ubound_unicode_version=unicode_version,
        hex_str_vs=HEX_STR_VS15)

    # Parse and join the earlier-format emoji release
    table[unicode_version].values.update(
        parse_vs_data(fname=UnicodeDataFile.LegacyEmojiVariationSequences(),
                      ubound_unicode_version=unicode_version,
                      hex_str_vs=HEX_STR_VS15).values)

    # Keep only characters that are already wide (width 2) -- these are the
    # ones where VS15 has a narrowing effect. Narrow characters (width 1)
    # stay narrow regardless.
    wide_table = wide_tables[unicode_latest].as_value_ranges()
    table[unicode_version].values = {
        ucs for ucs in table[unicode_version].values
        if _bisearch(ucs, wide_table)
    }

    return UnicodeTableRenderCtx('VS15_WIDE_TO_NARROW', table)


def parse_vs_data(fname: str, ubound_unicode_version: UnicodeVersion, hex_str_vs: str):
    with open(fname, encoding='utf-8') as fin:
        table_iter = parse_vs_table(fin, hex_str_vs)
        # pull "date string"
        date = next(table_iter).comment.split(':', 1)[1].strip()
        # pull values only matching this unicode version and lower
        values = {entry.code_range[0] for entry in table_iter}
    return TableDef(ubound_unicode_version, date, values)


def cite_source_description(filename: str) -> tuple[str, str]:
    """Return unicode.org source data file's own description as citation."""
    with open(filename, encoding='utf-8') as f:
        entry_iter = parse_unicode_table(f)
        fname = next(entry_iter).comment.strip()
        # use local name w/version in place of 'emoji-variation-sequences.txt'
        if fname == 'emoji-variation-sequences.txt':
            fname = os.path.basename(filename)
        date = next(entry_iter).comment.strip()

    return fname, date


def name_ucs(ucs: str) -> str:
    try:
        return string.capwords(unicodedata.name(ucs))
    except ValueError:
        return None


def parse_unicode_table(file: Iterable[str]) -> Iterator[TableEntry]:
    """
    Parse unicode tables.

    See details: https://www.unicode.org/reports/tr44/#Format_Conventions
    """
    for line in file:
        data, _, comment = line.partition('#')
        data_fields: Iterator[str] = (field.strip() for field in data.split(';'))
        code_points_str, *properties = data_fields

        if not code_points_str:
            yield TableEntry(None, tuple(properties), comment)
            continue

        if '..' in code_points_str:
            start, end = code_points_str.split('..')
        else:
            start = end = code_points_str
        code_range = (int(start, base=16), int(end, base=16) + 1)

        yield TableEntry(code_range, tuple(properties), comment)


def parse_vs_table(fp: Iterable[str], hex_str_vs: str = 'FE0F') -> Iterator[TableEntry]:
    """Parse emoji-variation-sequences.txt for codepoints that precede `hex_str_vs`."""
    for line in fp:
        data, _, comment = line.partition('#')
        data_fields: Iterator[str] = (field.strip() for field in data.split(';'))
        code_points_str, *properties = data_fields

        if not code_points_str:
            if 'Date' in comment:
                # yield 'Data'
                yield TableEntry(None, tuple(properties), comment)
            continue
        code_points = code_points_str.split()
        if len(code_points) == 2 and code_points[1] == hex_str_vs:
            # yield a single "code range" entry for a single value that precedes hex_str_vs
            yield TableEntry((int(code_points[0], 16), int(code_points[0], 16)), tuple(properties), comment)


@functools.cache
def parse_category(fname: str, wide: int) -> TableDef:
    """Parse value ranges of unicode data files, by given categories into string tables."""
    print(f'parsing {fname}, wide={wide}: ', end='', flush=True)

    with open(fname, encoding='utf-8') as f:
        table_iter = parse_unicode_table(f)

        # pull "version string" from first line of source file
        version = next(table_iter).comment.strip()
        # and "date string" from second line
        date = next(table_iter).comment.split(':', 1)[1].strip()
        values = TableEntry.parse_width_category_values(table_iter, wide)
    print('ok')
    return TableDef(version, date, values)


@functools.cache
def parse_category_ambiguous(fname: str) -> TableDef:
    """Parse EastAsianWidth.txt for 'A' (Ambiguous) category."""
    print(f'parsing {fname}, category=A: ', end='', flush=True)

    with open(fname, encoding='utf-8') as f:
        table_iter = parse_unicode_table(f)

        # pull "version string" from first line of source file
        version = next(table_iter).comment.strip()
        # and "date string" from second line
        date = next(table_iter).comment.split(':', 1)[1].strip()
        values = {
            n
            for entry in table_iter
            if entry.code_range is not None and entry.properties[0] == 'A'
            for n in range(entry.code_range[0], entry.code_range[1])
        }
    print('ok')
    return TableDef(version, date, values)


def parse_grapheme_break_properties(fname: str) -> dict[str, TableDef]:
    """Parse GraphemeBreakProperty.txt for grapheme break properties needing tables."""
    print(f'parsing {fname}: ', end='', flush=True)
    values_by_prop: dict[str, set[int]] = {prop: set() for prop in GRAPHEME_BREAK_PROPERTIES}

    with open(fname, encoding='utf-8') as f:
        table_iter = parse_unicode_table(f)
        version = next(table_iter).comment.strip()
        date = next(table_iter).comment.split(':', 1)[1].strip()

        for entry in table_iter:
            if entry.code_range is None:
                continue
            if entry.properties and entry.properties[0] in values_by_prop:
                values_by_prop[entry.properties[0]].update(
                    range(entry.code_range[0], entry.code_range[1])
                )

    print('ok')
    return {
        f'GRAPHEME_{prop.upper()}': TableDef(version, date, values)
        for prop, values in values_by_prop.items()
    }


def parse_extended_pictographic(fname: str) -> TableDef:
    """Parse emoji-data.txt for Extended_Pictographic property."""
    print(f'parsing {fname} for Extended_Pictographic: ', end='', flush=True)
    values: set[int] = set()

    with open(fname, encoding='utf-8') as f:
        table_iter = parse_unicode_table(f)
        # pull "version string" from first line of source file
        version = next(table_iter).comment.strip()
        # and "date string" from second line
        date = next(table_iter).comment.split(':', 1)[1].strip()

        for entry in table_iter:
            if entry.code_range is None:
                continue
            if entry.properties and entry.properties[0] == 'Extended_Pictographic':
                values.update(range(entry.code_range[0], entry.code_range[1]))

    print('ok')
    return TableDef(version, date, values)


def parse_indic_conjunct_breaks(fname: str) -> dict[str, TableDef]:
    """Parse DerivedCoreProperties.txt for all Indic_Conjunct_Break properties."""
    print(f'parsing {fname} for InCB: ', end='', flush=True)
    values_by_incb: dict[str, set[int]] = {val: set() for val in INCB_VALUES}

    with open(fname, encoding='utf-8') as f:
        for line in f:
            data, _, comment = line.partition('#')
            data = data.strip()
            if not data:
                continue

            parts = [p.strip() for p in data.split(';')]
            if len(parts) < 3:
                continue

            code_points_str, prop_name, prop_value = parts[0], parts[1], parts[2]

            if prop_name == 'InCB' and prop_value in values_by_incb:
                if '..' in code_points_str:
                    start, end = code_points_str.split('..')
                    values_by_incb[prop_value].update(
                        range(int(start, 16), int(end, 16) + 1)
                    )
                else:
                    values_by_incb[prop_value].add(int(code_points_str, 16))

    print('ok')
    return {
        f'INCB_{val.upper()}': TableDef('DerivedCoreProperties', 'see file', values)
        for val, values in values_by_incb.items()
    }


# ISC_CONSONANT is generated for the public API but not used by the runtime.
# Virama and Invisible_Stacker are treated as zero-width combining marks (Mn),
# matching Ghostty's uucode; the runtime no longer consults the ISC categories.
ISC_VALUES = ('Consonant', 'Virama', 'Invisible_Stacker')


def parse_indic_syllabic_category(fname: str) -> dict[str, TableDef]:
    """
    Parse IndicSyllabicCategory.txt for Consonant property.

    See https://www.unicode.org/reports/tr44/#Indic_Syllabic_Category
    """
    print(f'parsing {fname} for ISC: ', end='', flush=True)
    values_by_isc: dict[str, set[int]] = {val: set() for val in ISC_VALUES}

    with open(fname, encoding='utf-8') as f:
        for line in f:
            data, _, comment = line.partition('#')
            data = data.strip()
            if not data:
                continue

            parts = [p.strip() for p in data.split(';')]
            if len(parts) < 2:
                continue

            code_points_str, prop_value = parts[0], parts[1]

            if prop_value in values_by_isc:
                if '..' in code_points_str:
                    start, end = code_points_str.split('..')
                    values_by_isc[prop_value].update(
                        range(int(start, 16), int(end, 16) + 1)
                    )
                else:
                    values_by_isc[prop_value].add(int(code_points_str, 16))

    print('ok')
    return {
        f'ISC_{val.upper()}': TableDef('IndicSyllabicCategory', 'see file', values)
        for val, values in values_by_isc.items()
    }


@functools.lru_cache(maxsize=None)
def parse_derived_core_property(fname: str, property_name: str) -> set[int]:
    """Parse DerivedCoreProperties.txt for a specific property."""
    print(f'parsing {fname} for {property_name}: ', end='', flush=True)
    values: set[int] = set()

    with open(fname, encoding='utf-8') as f:
        for line in f:
            data, _, comment = line.partition('#')
            data = data.strip()
            if not data:
                continue

            parts = [p.strip() for p in data.split(';')]
            if len(parts) < 2:
                continue

            code_points_str, prop_name = parts[0], parts[1]

            if prop_name == property_name:
                if '..' in code_points_str:
                    start, end = code_points_str.split('..')
                    values.update(range(int(start, 16), int(end, 16) + 1))
                else:
                    values.add(int(code_points_str, 16))

    print('ok')
    return values


def fetch_table_grapheme_data() -> GraphemeTableRenderCtx:
    """Fetch grapheme break property tables for the latest Unicode version only."""
    latest_version = fetch_unicode_versions()[-1]

    # makes a table definition for each break property
    tables = parse_grapheme_break_properties(
        UnicodeDataFile.GraphemeBreakProperty(latest_version)
    )
    tables['EXTENDED_PICTOGRAPHIC'] = parse_extended_pictographic(
        UnicodeDataFile.EmojiData(latest_version)
    )
    tables.update(parse_indic_conjunct_breaks(
        UnicodeDataFile.DerivedCoreProperties(latest_version)
    ))
    tables.update(parse_indic_syllabic_category(
        UnicodeDataFile.IndicSyllabicCategory(latest_version)
    ))
    return GraphemeTableRenderCtx(str(latest_version), tables)


class UnicodeDataFile:
    """
    Helper class for fetching Unicode Data Files.

    Methods like 'DerivedAge' return a local filename, but have the side-effect of fetching those
    files from unicode.org first, if not existing or out-of-date.

    Because file modification times are used, for local files of TestEmojiZWJSequences and
    TestEmojiVariationSequences, these files should be forcefully re-fetched CLI argument '--no-
    check-last-modified'.
    """

    URL_DERIVED_AGE = 'https://www.unicode.org/Public/UCD/latest/ucd/DerivedAge.txt'
    URL_EASTASIAN_WIDTH = 'https://www.unicode.org/Public/{version}/ucd/EastAsianWidth.txt'
    URL_DERIVED_CATEGORY = 'https://www.unicode.org/Public/{version}/ucd/extracted/DerivedGeneralCategory.txt'
    URL_EMOJI_VARIATION = 'https://unicode.org/Public/{version}/ucd/emoji/emoji-variation-sequences.txt'
    URL_LEGACY_VARIATION = 'https://unicode.org/Public/emoji/{version}/emoji-variation-sequences.txt'
    URL_EMOJI_ZWJ = 'https://unicode.org/Public/emoji/{version}/emoji-zwj-sequences.txt'
    URL_GRAPHEME_BREAK = 'https://www.unicode.org/Public/{version}/ucd/auxiliary/GraphemeBreakProperty.txt'
    URL_EMOJI_DATA = 'https://www.unicode.org/Public/{version}/ucd/emoji/emoji-data.txt'
    URL_DERIVED_CORE_PROPS = 'https://www.unicode.org/Public/{version}/ucd/DerivedCoreProperties.txt'
    URL_PROP_LIST = 'https://www.unicode.org/Public/{version}/ucd/PropList.txt'
    URL_GRAPHEME_BREAK_TEST = 'https://www.unicode.org/Public/{version}/ucd/auxiliary/GraphemeBreakTest.txt'
    URL_INDIC_SYLLABIC_CATEGORY = 'https://www.unicode.org/Public/{version}/ucd/IndicSyllabicCategory.txt'
    URL_UDHR_ZIP = 'http://efele.net/udhr/assemblies/udhr_txt.zip'

    @classmethod
    def DerivedAge(cls) -> str:
        fname = os.path.join(PATH_DATA, 'DerivedAge.txt')
        cls.do_retrieve(url=cls.URL_DERIVED_AGE, fname=fname)
        return fname

    @classmethod
    def EastAsianWidth(cls, version: str) -> str:
        fname = os.path.join(PATH_DATA, f'EastAsianWidth-{version}.txt')
        cls.do_retrieve(url=cls.URL_EASTASIAN_WIDTH.format(version=version), fname=fname)
        return fname

    @classmethod
    def DerivedGeneralCategory(cls, version: str) -> str:
        fname = os.path.join(PATH_DATA, f'DerivedGeneralCategory-{version}.txt')
        cls.do_retrieve(url=cls.URL_DERIVED_CATEGORY.format(version=version), fname=fname)
        return fname

    @classmethod
    def EmojiVariationSequences(cls, version: str) -> str:
        fname = os.path.join(PATH_DATA, f'emoji-variation-sequences-{version}.txt')
        cls.do_retrieve(url=cls.URL_EMOJI_VARIATION.format(version=version), fname=fname)
        return fname

    @classmethod
    def LegacyEmojiVariationSequences(cls) -> str:
        version = "12.0"
        fname = os.path.join(PATH_DATA, f'emoji-variation-sequences-{version}.0.txt')
        cls.do_retrieve(url=cls.URL_LEGACY_VARIATION.format(version=version), fname=fname)
        return fname

    @classmethod
    def TestEmojiVariationSequences(cls) -> str:
        version = fetch_unicode_versions()[-1]
        fname = os.path.join(PATH_TESTS, 'emoji-variation-sequences.txt')
        cls.do_retrieve(url=cls.URL_EMOJI_VARIATION.format(version=version), fname=fname)
        return fname

    @classmethod
    def TestEmojiZWJSequences(cls) -> str:
        # ZWJ sequences are only at /Public/emoji/{version}/, use 'latest' for tests
        fname = os.path.join(PATH_TESTS, 'emoji-zwj-sequences.txt')
        cls.do_retrieve(url=cls.URL_EMOJI_ZWJ.format(version='latest'), fname=fname)
        return fname

    @classmethod
    def GraphemeBreakProperty(cls, version: str) -> str:
        fname = os.path.join(PATH_DATA, f'GraphemeBreakProperty-{version}.txt')
        cls.do_retrieve(url=cls.URL_GRAPHEME_BREAK.format(version=version), fname=fname)
        return fname

    @classmethod
    def EmojiData(cls, version: UnicodeVersion) -> str:
        """Fetch emoji-data.txt for Extended_Pictographic property."""
        fname = os.path.join(PATH_DATA, f'emoji-data-{version}.txt')
        cls.do_retrieve(url=cls.URL_EMOJI_DATA.format(version=version), fname=fname)
        return fname

    @classmethod
    def DerivedCoreProperties(cls, version: str) -> str:
        fname = os.path.join(PATH_DATA, f'DerivedCoreProperties-{version}.txt')
        cls.do_retrieve(url=cls.URL_DERIVED_CORE_PROPS.format(version=version), fname=fname)
        return fname

    @classmethod
    def PropList(cls, version: str) -> str:
        fname = os.path.join(PATH_DATA, f'PropList-{version}.txt')
        cls.do_retrieve(url=cls.URL_PROP_LIST.format(version=version), fname=fname)
        return fname

    @classmethod
    def IndicSyllabicCategory(cls, version: str) -> str:
        fname = os.path.join(PATH_DATA, f'IndicSyllabicCategory-{version}.txt')
        cls.do_retrieve(url=cls.URL_INDIC_SYLLABIC_CATEGORY.format(version=version), fname=fname)
        return fname

    @classmethod
    def TestGraphemeBreakTest(cls) -> str:
        version = fetch_unicode_versions()[-1]
        fname = os.path.join(PATH_TESTS, 'GraphemeBreakTest.txt')
        cls.do_retrieve(url=cls.URL_GRAPHEME_BREAK_TEST.format(version=version), fname=fname)
        return fname

    @classmethod
    def UDHRCombined(cls) -> str:
        """
        Fetch UDHR zip, extract and combine all translations into a single file.

        Downloads http://efele.net/udhr/assemblies/udhr_txt.zip, extracts the text files,
        and combines them with '--' separator between translations.
        """
        fname = os.path.join(PATH_TESTS, 'udhr_combined.txt')
        cls.do_retrieve_udhr_combined(url=cls.URL_UDHR_ZIP, fname=fname)
        return fname

    @staticmethod
    def do_retrieve_udhr_combined(url: str, fname: str) -> None:
        """Fetch UDHR zip file, extract, and combine all translations."""
        if not UnicodeDataFile.is_url_newer(url, fname):
            return

        session = UnicodeDataFile.get_http_session()
        print(f"fetching {url}: ", end='', flush=True)
        resp = session.get(url, timeout=(CONNECT_TIMEOUT, READ_TIMEOUT))
        resp.raise_for_status()
        print('ok')

        print(f"extracting and combining to {fname}: ", end='', flush=True)
        combined_content = []
        with zipfile.ZipFile(io.BytesIO(resp.content)) as zf:
            txt_files = sorted([n for n in zf.namelist() if n.endswith('.txt')])
            for txt_name in txt_files:
                with zf.open(txt_name) as f:
                    content = f.read().decode('utf-8').rstrip()
                    combined_content.append(f'----\n\n{content}\n')

        folder = os.path.dirname(fname)
        if folder and not os.path.exists(folder):
            os.makedirs(folder)

        with open(fname, 'w', encoding='utf-8', newline='\n') as fout:
            fout.writelines(combined_content)
        print('ok')

    @staticmethod
    def do_retrieve(url: str, fname: str) -> None:
        """Retrieve given url to target filepath fname."""
        folder = os.path.dirname(fname)
        if folder and not os.path.exists(folder):
            os.makedirs(folder)
        if not UnicodeDataFile.is_url_newer(url, fname):
            return
        session = UnicodeDataFile.get_http_session()
        resp = session.get(url, timeout=(CONNECT_TIMEOUT, READ_TIMEOUT))
        resp.raise_for_status()
        print(f"saving {fname}: ", end='', flush=True)
        with open(fname, 'wb') as fout:
            for chunk in resp.iter_content(FETCH_BLOCKSIZE):
                fout.write(chunk)
        print('ok')

    @staticmethod
    def is_url_newer(url: str, fname: str) -> bool:
        if not os.path.exists(fname):
            return True
        if CHECK_LAST_MODIFIED:
            session = UnicodeDataFile.get_http_session()
            resp = session.head(url, timeout=(CONNECT_TIMEOUT, READ_TIMEOUT))
            resp.raise_for_status()
            remote_url_dt = dateutil.parser.parse(resp.headers['Last-Modified']).astimezone()
            local_file_dt = datetime.datetime.fromtimestamp(os.path.getmtime(fname)).astimezone()
            return remote_url_dt > local_file_dt
        return False

    @functools.cache
    def get_http_session() -> requests.Session:
        session = requests.Session()
        session.headers.update({
            'User-Agent': 'wcwidth-update-tables/1.0 (https://github.com/jquast/wcwidth)'
        })
        retries = urllib3.util.Retry(
            total=MAX_RETRIES,
            connect=MAX_RETRIES,
            read=MAX_RETRIES,
            backoff_factor=BACKOFF_FACTOR,
            backoff_jitter=BACKOFF_FACTOR,
            status_forcelist=[408, 429, 500, 502, 503, 504, 520],
            allowed_methods=['HEAD', 'GET'],
            raise_on_status=False,
            respect_retry_after_header=True,
        )
        adapter = requests.adapters.HTTPAdapter(max_retries=retries)
        session.mount('https://', adapter)
        session.mount('http://', adapter)
        return session

    @staticmethod
    def filenames() -> list[str]:
        """Return list of UnicodeData files stored in PATH_DATA, sorted by version number."""
        pattern = re.compile(
            r'^(emoji-variation-sequences|DerivedGeneralCategory|EastAsianWidth)-(\d+)\.(\d+)\.(\d+).txt$')
        filename_matches = []
        for fname in os.listdir(PATH_DATA):
            if match := re.search(pattern, fname):
                filename_matches.append(match)
        filename_matches.sort(key=lambda m: (
            m.group(1),
            int(m.group(2)),
            int(m.group(3)),
            int(m.group(4)),
        ))
        return [os.path.join(PATH_DATA, match.string) for match in filename_matches]


def update_readme_term_programs() -> bool:
    """
    Update the ``list_term_programs()`` example in ``README.rst``.

    The section between ``.. BEGIN_LIST_TERM_PROGRAMS`` and ``.. END_LIST_TERM_PROGRAMS``
    is replaced with the current sorted terminal names (first 5 followed by ``...``).

    Returns True if the file was modified.
    """
    readme_path = os.path.join(PATH_UP, 'README.rst')
    with open(readme_path, encoding='utf-8') as fin:
        original = fin.read()

    tp = collect_term_programs()
    all_names = sorted(tp.known_terminals | tp.aliases.keys())
    display = textwrap.fill(repr(tuple(all_names)), width=79,
                            subsequent_indent='     ',
                            break_on_hyphens=False)

    output_lines = [
        '.. BEGIN_LIST_TERM_PROGRAMS',
        '.. code-block:: python',
        '',
        '    >>> wcwidth.list_term_programs()',
        f'    {display}',
        '',
        '.. END_LIST_TERM_PROGRAMS',
    ]
    replacement = '\n'.join(output_lines)

    pattern = re.compile(
        r'\.\. BEGIN_LIST_TERM_PROGRAMS\n.*?\n\.\. END_LIST_TERM_PROGRAMS',
        re.DOTALL,
    )
    modified = pattern.sub(replacement, original)

    if modified != original:
        with open(readme_path, 'w', encoding='utf-8', newline='\n') as fout:
            fout.write(modified)
        return True
    return False


# These appear to share the same engines.
SOFTWARE_SHARED_ENGINES = {
    'QTerminal': 'qtermwidget',
    'cool-retro-term': 'qtermwidget',
    'Hyper': 'xterm.js',
    'Tabby': 'xterm.js',
    'st-luke': 'st',
}

VTE_CANONICAL = 'vte'


def canonical_name(software_name: str, software_version: str) -> str:
    """Determine the canonical terminal name, applying VTE and known consolidations."""
    if 'VTE' in software_version:
        return VTE_CANONICAL.lower()
    return SOFTWARE_SHARED_ENGINES.get(software_name, software_name).lower()


def parse_wchar_codepoint(wchar: str) -> int:
    r"""
    Extract the primary codepoint from a wchar string.

    The ``wchar`` field in ucs-detect YAML is stored as a literal Python escape string
    (e.g. ``'\\u2630'``, ``'\\U0001f468'``), not as the actual Unicode character.
    Decode it first, then return the ordinal of the first character.

    For VS16/VS15 sequences like ``'\\u231a\\ufe0e'``, returns the base codepoint (0x231A).
    """
    decoded = wchar.encode('ascii').decode('unicode_escape')
    if len(decoded) > 1 and decoded[-1] in ('\ufe0f', '\ufe0e'):
        return ord(decoded[0])
    return ord(decoded)


@dataclass(frozen=True)
class TerminalOverrides:
    """Per-terminal codepoint override ranges for narrower/wider/zeroer measurements."""
    narrower: list[tuple[str, str, str]]
    wider: list[tuple[str, str, str]]
    zeroer: list[tuple[str, str, str]]
    narrow_zeroer: list[tuple[str, str, str]]

    def items(self) -> list[tuple[str, list[tuple[str, str, str]]]]:
        return [('narrower', self.narrower),
                ('wider', self.wider),
                ('zeroer', self.zeroer),
                ('narrow_zeroer', self.narrow_zeroer)]


@dataclass(frozen=True)
class DedupedOverrides:
    """Deduplicated override data: shared sets keyed by hash, terminal-to-hash refs."""
    shared_sets: dict[str, TerminalOverrides]
    terminal_refs: dict[str, str]


def dedup_override_table(
    table: Mapping[str, TerminalOverrides],
) -> DedupedOverrides:
    """Deduplicate override table by hashing."""
    shared_sets: dict[str, TerminalOverrides] = {}
    terminal_refs: dict[str, str] = {}
    for term_name, overrides in table.items():
        key = (tuple(overrides.narrower),
               tuple(overrides.wider),
               tuple(overrides.zeroer),
               tuple(overrides.narrow_zeroer))
        hash_key = hashlib.sha256(repr(key).encode()).hexdigest()[:8]
        if hash_key not in shared_sets:
            shared_sets[hash_key] = overrides
        terminal_refs[term_name] = hash_key
    return DedupedOverrides(shared_sets=shared_sets, terminal_refs=terminal_refs)


@dataclass(frozen=True)
class MergedOverridesCategory:
    """A single category within the merged overrides table."""
    variable_name: str
    shared_sets: Mapping[str, TerminalOverrides]
    terminal_refs: Mapping[str, str]
    set_terminals: Mapping[str, tuple[str, ...]]


@dataclass(frozen=True)
class MergedOverridesRenderCtx(RenderContext):
    """Render context for all single-codepoint override tables in one file."""
    categories: Sequence[MergedOverridesCategory]


@dataclass
class MergedOverridesRenderDef(RenderDefinition):
    render_context: MergedOverridesRenderCtx

    @classmethod
    def new(cls, categories: Sequence[MergedOverridesCategory]) -> Self:
        return cls(
            jinja_filename='table_overrides.py.j2',
            output_filename=os.path.join(PATH_UP, 'wcwidth', 'table_overrides.py'),
            render_context=MergedOverridesRenderCtx(categories=categories),
        )


@dataclass(frozen=True)
class GraphemeOverridePerTerminalRenderCtx(RenderContext):
    """Render context for a single terminal's grapheme overrides."""
    canonical_name: str
    graphemes: dict[str, int]
    terminals: list[str]      # All terminals that share this same grapheme data


@dataclass
class GraphemeOverridePerTerminalRenderDef(RenderDefinition):
    render_context: GraphemeOverridePerTerminalRenderCtx

    @classmethod
    def new(cls, canonical_name: str, graphemes: dict[str, int],
            terminals: list[str]) -> Self:
        safe_name = canonical_name.replace('-', '_').replace('.', '_')
        filename = f'table_grapheme_overrides/{safe_name}.py'
        return cls(
            jinja_filename='grapheme_override_per_terminal.py.j2',
            output_filename=os.path.join(PATH_UP, 'wcwidth', filename),
            render_context=GraphemeOverridePerTerminalRenderCtx(
                canonical_name=canonical_name,
                graphemes=graphemes,
                terminals=terminals,
            ),
        )


@dataclass(frozen=True)
class GraphemeRegistryRenderCtx(RenderContext):
    """Render context for the grapheme override hash registry."""

    registry: dict[str, str]


@dataclass
class GraphemeRegistryRenderDef(RenderDefinition):
    render_context: GraphemeRegistryRenderCtx

    @classmethod
    def new(cls, registry: dict[str, str]) -> Self:
        return cls(
            jinja_filename='grapheme_registry.py.j2',
            output_filename=os.path.join(
                PATH_UP, 'wcwidth', 'table_grapheme_overrides', '_registry.py'),
            render_context=GraphemeRegistryRenderCtx(registry=registry),
        )


def values_to_hex_ranges(values: set[int]) -> list[tuple[str, str, str]]:
    """Convert a set of codepoint integers to hex range descriptions."""
    if not values:
        return []
    sorted_vals = sorted(values)
    ranges = []
    start = end = sorted_vals[0]
    for val in sorted_vals[1:]:
        if val == end + 1:
            end = val
        else:
            ranges.append((start, end))
            start = end = val
    ranges.append((start, end))

    result: list[tuple[str, str, str]] = []
    for lo, hi in ranges:
        hex_start, hex_end = f'0x{lo:05x}', f'0x{hi:05x}'
        name_start = name_ucs(chr(lo)) or '(nil)'
        name_end = name_ucs(chr(hi)) or '(nil)'
        if name_start != name_end:
            txt = f'{name_start[:24].rstrip():24s}..{name_end[:24].rstrip()}'
        else:
            txt = name_start[:48]
        result.append((hex_start, hex_end, txt))
    return result


@functools.lru_cache(maxsize=1)
def load_ucs_detect_yaml() -> list[tuple[str, str, Any]]:
    """Return (filename, canonical_name, yaml_document) for each ucs-detect data file."""
    items: list[tuple[str, str, Any]] = []
    for yaml_path in sorted(glob.glob(os.path.join(PATH_UCS_DETECT_DATA, '*.yaml'))):
        with open(yaml_path, encoding='utf-8') as f:
            doc = yaml.load(f, Loader=SafeLoader)
        name = doc.get('software_name', '')
        ver = doc.get('software_version', '')
        canonical = canonical_name(name, ver)
        items.append((os.path.basename(yaml_path), canonical, doc))
    return items


def make_single_override(
    category: str,
    known_terminals: frozenset[str],
) -> Mapping[str, TerminalOverrides]:
    """
    Collect single-codepoint overrides for a given test_results category.

    Returns a dict mapping canonical_name to 'TerminalOverrides'.

    - 'zeroer' means terminal measured 0, but wcwidth measured 2,
    - 'narrower' means terminal measured 1, but wcwidth measured 2,
    - 'wider' means terminal measured 2, but wcwidth measured 1.
    """
    zeroer: dict[str, set[int]] = {}
    narrower: dict[str, set[int]] = {}
    wider: dict[str, set[int]] = {}
    narrow_zeroer_map: dict[str, set[int]] = {}

    for _, canonical, doc in load_ucs_detect_yaml():
        test_results = doc.get('test_results', {})
        cat_results = test_results.get(category, {})
        for _ver, ver_data in cat_results.items():
            for entry in ver_data.get('failed_codepoints', []):
                wchar = entry['wchar']
                ucs = parse_wchar_codepoint(wchar)
                term_w = entry['measured_by_terminal']
                wc_w = entry['measured_by_wcwidth']
                if term_w == 0 and wc_w == 2:
                    zeroer.setdefault(canonical, set()).add(ucs)
                elif term_w == 0 and wc_w == 1 and category == 'narrow_results':
                    narrow_zeroer_map.setdefault(canonical, set()).add(ucs)
                elif term_w == 1 and wc_w == 2:
                    narrower.setdefault(canonical, set()).add(ucs)
                # 'wider' entries in emoji_vs16_results are from the vs16n baseline test
                # (base character measured without VS16, expected width 1).  Kitty rendering
                # these codepoints as 2 cells without VS16 is a WIDE_OVERRIDES issue, not
                # VS16.  Terminal multiplexer entries (e.g. tmux, libvterm) are excluded by
                # known_terminals filtering below.
                elif category != 'emoji_vs16_results' and term_w == 2 and wc_w == 1:
                    wider.setdefault(canonical, set()).add(ucs)

    result: dict[str, TerminalOverrides] = {}
    all_names = sorted(set(zeroer.keys()) | set(narrower.keys()) | set(wider.keys())
                       | set(narrow_zeroer_map.keys()))
    for name in all_names:
        result[name] = TerminalOverrides(
            narrower=values_to_hex_ranges(narrower.get(name, set())),
            wider=values_to_hex_ranges(wider.get(name, set())),
            zeroer=values_to_hex_ranges(zeroer.get(name, set())),
            narrow_zeroer=values_to_hex_ranges(narrow_zeroer_map.get(name, set())),
        )
    result = {
        name: data for name, data in result.items()
        if name in known_terminals
    }
    return result


def collect_grapheme_overrides(
    known_terminals: frozenset[str],
) -> Mapping[str, dict[str, int]]:
    """
    Collect multi-codepoint grapheme overrides from emoji_zwj_results, ri_results and
    language_results.

    Returns a dict mapping canonical_name -> {grapheme_string: terminal_measured_width}. Only
    includes entries where the terminal measurement differs from wcwidth. Grapheme strings are
    stored as Python source string literals suitable for code generation.
    """
    result: dict[str, dict[str, int]] = {}
    categories = ('emoji_zwj_results', 'ri_results')

    for _, canonical, doc in load_ucs_detect_yaml():
        test_results = doc.get('test_results', {})
        term_graphemes: dict[str, int] = {}
        for category in categories:
            cat_results = test_results.get(category, {})
            for _ver, ver_data in cat_results.items():
                for entry in ver_data.get('failed_codepoints', []):
                    wchar = entry['wchar']
                    term_w = entry['measured_by_terminal']
                    wc_w = entry['measured_by_wcwidth']
                    if term_w != wc_w:
                        decoded = wchar.encode('ascii').decode('unicode_escape')
                        term_graphemes[decoded] = term_w

        lang_results = test_results.get('language_results')
        if lang_results:
            for lang_data in lang_results.values():
                if not isinstance(lang_data, dict):
                    continue
                for entry in lang_data.get('failed', []):
                    if 'inherited_from' in entry:
                        continue
                    wchar = entry['wchars']
                    term_w = entry['measured_by_terminal']
                    wc_w = entry['measured_by_wcwidth']
                    if term_w != wc_w:
                        decoded = wchar.encode('ascii').decode('unicode_escape')
                        term_graphemes[decoded] = term_w

        if term_graphemes:
            result.setdefault(canonical, {}).update(term_graphemes)

    result = {
        name: data for name, data in result.items()
        if name in known_terminals
    }
    return result


def _make_merged_category(variable_name: str,
                          table: Mapping[str, TerminalOverrides]) -> MergedOverridesCategory:
    deduped = dedup_override_table(table)
    set_terminals: dict[str, tuple[str, ...]] = {}
    for term_name, hash_key in deduped.terminal_refs.items():
        set_terminals.setdefault(hash_key, []).append(term_name)
    set_terminals = {k: tuple(sorted(v)) for k, v in set_terminals.items()}
    return MergedOverridesCategory(variable_name, deduped.shared_sets,
                                   deduped.terminal_refs, set_terminals)


def fetch_override_grapheme_data(known_terminals: frozenset[str]) -> list[RenderDefinition]:
    """Generate shared GRAPHEME_OVERRIDES files, deduplicating identical tables."""
    table = collect_grapheme_overrides(known_terminals)

    # Group terminals by stable hash of their grapheme data
    hash_groups: dict[str, list[str]] = {}
    terminal_hashes: dict[str, str] = {}

    for canonical_name, graphemes in sorted(table.items()):
        if not graphemes:
            continue
        sorted_items = tuple(sorted(graphemes.items()))
        hash_key = hashlib.sha256(repr(sorted_items).encode()).hexdigest()[:8]
        hash_groups.setdefault(hash_key, []).append(canonical_name)
        terminal_hashes[canonical_name] = hash_key

    result: list[RenderDefinition] = []

    # Generate one shared file per unique hash
    for hash_key, terminals in sorted(hash_groups.items()):
        graphemes = table[terminals[0]]
        shared_name = f'_known_{hash_key}'
        result.append(
            GraphemeOverridePerTerminalRenderDef.new(shared_name, graphemes, terminals))

    # Generate registry mapping terminal -> hash
    result.append(GraphemeRegistryRenderDef.new(terminal_hashes))

    return result


@dataclass(frozen=True)
class TermPrograms:
    """Canonical terminal names and aliases collected from ucs-detect data."""
    known_terminals: frozenset[str]
    aliases: dict[str, str]


@functools.lru_cache(maxsize=1)
def collect_term_programs() -> TermPrograms:
    """Collect canonical terminal names and aliases from ucs-detect data."""
    # Only terminals detectable by XTVERSION, ENQ, TERM_PROGRAM, or unique TERM are included.
    known: set[str] = set()
    tprog_aliases: dict[str, str] = {}
    term_aliases: dict[str, str] = {}

    for _, canonical, doc in load_ucs_detect_yaml():
        tr = doc.get('terminal_results') or {}
        method = tr.get('software_method', '') or ''
        env = doc.get('environment') or {}
        tprog = (env.get('TERM_PROGRAM', '') or '').strip()
        term = (env.get('TERM', '') or '').strip()
        ver = doc.get('software_version', '') or ''

        has_xtversion = method in ('XTVERSION', 'ENQ') or 'VTE' in ver
        has_tprog = bool(tprog)
        has_dist_term = bool(term and term.lower() not in ('xterm-256color', 'xterm'))

        if not (has_xtversion or has_tprog or has_dist_term):
            print(f'terminal "{canonical}" ({ver}) is not auto-detectable', file=sys.stderr)
            continue

        known.add(canonical)

        if has_tprog:
            key = tprog.lower()
            if key and key != canonical:
                tprog_aliases[key] = canonical

        if has_dist_term:
            key = term.lower()
            if key != canonical:
                term_aliases.setdefault(key, canonical)

    # aliases for well-known TERM_PROGRAM values not in ucs-detect data.
    tprog_aliases.update({
        'rxvt': 'urxvt',
        'vscode': 'xterm.js',
    })

    # hardcoded aliases for well-known TERM values not in ucs-detect data.
    term_aliases.update({
        'putty': 'pterm',
    })

    # Terminal multiplexers (subterminals) depend on a host terminal for
    # rendering; their cursor-position reports from ucs-detect are not
    # reliable indicators of display width.
    _multiplexers = frozenset({'gnu screen', 'libvterm', 'tmux', 'zellij'})
    known -= _multiplexers
    term_aliases = {k: v for k, v in term_aliases.items() if v not in _multiplexers}
    tprog_aliases = {k: v for k, v in tprog_aliases.items() if v not in _multiplexers}

    return TermPrograms(
        known_terminals=frozenset(known),
        aliases={**term_aliases, **tprog_aliases},
    )


@dataclass(frozen=True)
class TermProgramTableRenderCtx(RenderContext):
    """Render context for terminal program data."""
    known_terminals: frozenset[str]
    aliases: dict[str, str]


@dataclass
class TermProgramTableRenderDef(RenderDefinition):
    render_context: TermProgramTableRenderCtx

    @classmethod
    def new(cls) -> Self:
        tp = collect_term_programs()
        return cls(
            jinja_filename='term_programs.py.j2',
            output_filename=os.path.join(PATH_UP, 'wcwidth', 'table_term_programs.py'),
            render_context=TermProgramTableRenderCtx(
                known_terminals=tp.known_terminals,
                aliases=tp.aliases,
            ),
        )


def fetch_all_emoji_files() -> None:
    """
    Fetch emoji variation sequences and ZWJ sequences for all versions.

    URL locations:
    - Variation sequences (5.0-12.1): /Public/emoji/{version}/
    - Variation sequences (13.0+): /Public/{version}/ucd/emoji/
    - ZWJ sequences (ALL versions): /Public/emoji/{version}/

    Note: ZWJ files never moved to /Public/{version}/ucd/emoji/ - they remain
    at /Public/emoji/{version}/ for all emoji versions.
    """
    unicode_versions = fetch_unicode_versions()

    # Legacy variation sequences (before 13.0, at /Public/emoji/{version}/)
    legacy_variation_versions = ['5.0', '11.0', '12.0', '12.1']

    for emoji_version in legacy_variation_versions:
        fname = os.path.join(PATH_DATA, f'emoji-variation-sequences-emoji-{emoji_version}.txt')
        UnicodeDataFile.do_retrieve(
            url=UnicodeDataFile.URL_LEGACY_VARIATION.format(version=emoji_version),
            fname=fname)

    # ZWJ sequences are ALL at /Public/emoji/{version}/ (they didn't move)
    all_zwj_versions = ['5.0', '11.0', '12.0', '12.1', '13.0', '13.1',
                        '14.0', '15.0', '15.1', '16.0']

    for emoji_version in all_zwj_versions:
        fname = os.path.join(PATH_DATA, f'emoji-zwj-sequences-emoji-{emoji_version}.txt')
        UnicodeDataFile.do_retrieve(
            url=UnicodeDataFile.URL_EMOJI_ZWJ.format(version=emoji_version),
            fname=fname)

    # Starting with Unicode 13.0.0, variation sequences moved to /Public/{version}/ucd/emoji/
    for version in unicode_versions:
        if version >= UnicodeVersion.parse('13.0.0'):
            fname = os.path.join(PATH_DATA, f'emoji-variation-sequences-{version}.txt')
            UnicodeDataFile.do_retrieve(
                url=UnicodeDataFile.URL_EMOJI_VARIATION.format(version=version),
                fname=fname)


def parse_args() -> dict[str, Any]:
    """Parse command-line arguments."""
    parser = argparse.ArgumentParser(
        description='Update Unicode code tables for wcwidth using jinja2 code generation.',
        epilog='https://github.com/jquast/wcwidth'
    )
    parser.add_argument(
        '--only-fetch',
        action='store_true',
        help='Only fetch data files without processing or code generation'
    )
    parser.add_argument(
        '--fetch-all-versions',
        action='store_true',
        help='Fetch emoji variation sequences and ZWJ sequences for all Unicode versions '
             '(for archival/testing purposes)'
    )
    parser.add_argument(
        '--check-last-modified',
        action='store_true',
        help='Check if remote files are newer than local files (rarely needed)'
    )
    return vars(parser.parse_args())


def fetch_all_data_files(fetch_all_versions: bool = False) -> None:
    """
    Fetch all required Unicode data files.

    Fetches data files for code generation and test files. Files are only downloaded if they don't
    exist locally or if CHECK_LAST_MODIFIED is True and the remote file is newer.
    """
    # Fetch DerivedAge first to determine available Unicode versions
    UnicodeDataFile.DerivedAge()
    version = fetch_unicode_versions()[-1]

    # Fetch data files required for code generation
    UnicodeDataFile.EastAsianWidth(version)
    UnicodeDataFile.DerivedGeneralCategory(version)
    UnicodeDataFile.EmojiVariationSequences(version)
    UnicodeDataFile.LegacyEmojiVariationSequences()
    UnicodeDataFile.GraphemeBreakProperty(version)
    UnicodeDataFile.EmojiData(version)
    UnicodeDataFile.DerivedCoreProperties(version)
    UnicodeDataFile.PropList(version)
    UnicodeDataFile.IndicSyllabicCategory(version)

    # Fetch test data files
    UnicodeDataFile.TestEmojiVariationSequences()
    UnicodeDataFile.TestEmojiZWJSequences()
    UnicodeDataFile.TestGraphemeBreakTest()
    UnicodeDataFile.UDHRCombined()

    # Fetch all legacy emoji files if requested
    if fetch_all_versions:
        fetch_all_emoji_files()


def cleanup_stale_grapheme_files() -> None:
    """Remove stale per-terminal grapheme override files and unreferenced _known_* files."""
    overrides_dir = os.path.join(PATH_UP, 'wcwidth', 'table_grapheme_overrides')
    # Because hashes can change at any given next release, this is a big part of why we prefix our
    # files with '_' to indicate to "please do not try to import and use them!", at least not without
    # great care via _registry.py -- because they could disappear at any next release.
    #
    # Load registry to determine which _known_* files are still referenced
    registry: dict[str, str] = {}
    registry_path = os.path.join(overrides_dir, '_registry.py')
    if os.path.exists(registry_path):
        # std imports
        import importlib.util
        spec = importlib.util.spec_from_file_location('_registry', registry_path)
        if spec and spec.loader:
            mod = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(mod)
            registry = getattr(mod, '_REGISTRY', {})

    active_hashes = set(registry.values())

    for filename in sorted(os.listdir(overrides_dir)):
        if not filename.endswith('.py'):
            continue
        if filename in ('__init__.py', '_registry.py'):
            continue
        if filename.startswith('_known_'):
            # Remove _known_* files whose hash is not in the registry
            hash_key = filename[len('_known_'):-len('.py')]
            if hash_key not in active_hashes:
                filepath = os.path.join(overrides_dir, filename)
                os.unlink(filepath)
                print(f'removed unreferenced {filepath}')
            continue
        # Remove old per-terminal override files (now covered by shared _known_* files)
        filepath = os.path.join(overrides_dir, filename)
        os.unlink(filepath)
        print(f'removed obsolete {filepath}')


def main(only_fetch: bool = False, fetch_all_versions: bool = False,
         check_last_modified: bool = False) -> None:
    """Update east-asian, combining and zero width tables."""
    # Set global flag for HTTP requests to check Last-Modified headers
    global CHECK_LAST_MODIFIED
    CHECK_LAST_MODIFIED = check_last_modified

    # Always fetch data files first
    fetch_all_data_files(fetch_all_versions)

    # Exit early if only fetching was requested
    if only_fetch:
        print('Fetch complete (--only-fetch mode, skipping code generation)')
        return

    # This defines which jinja source templates map to which output filenames,
    # and what function defines the source data. We hope to add more source
    # language options using jinja2 templates, with minimal modification of the
    # code.
    def get_codegen_definitions() -> Iterator[RenderDefinition]:
        latest_version = fetch_unicode_versions()[-1]
        yield UnicodeVersionPyRenderDef.new(UnicodeVersionPyRenderCtx([latest_version]))
        yield UnicodeTableRenderDef.new('table_vs16.py', fetch_table_vs16_data())
        yield UnicodeTableRenderDef.new('table_vs15.py', fetch_table_vs15_data())
        yield UnicodeTableRenderDef.new('table_wide.py', fetch_table_wide_data())
        yield UnicodeTableRenderDef.new('table_zero.py', fetch_table_zero_data())
        yield UnicodeTableRenderDef.new('table_mc.py', fetch_table_category_mc_data())
        yield UnicodeTableRenderDef.new('table_ambiguous.py', fetch_table_ambiguous_data())
        yield GraphemeTableRenderDef.new(fetch_table_grapheme_data())
        yield UnicodeVersionRstRenderDef.new(fetch_source_headers())

        kt = collect_term_programs().known_terminals
        yield MergedOverridesRenderDef.new([
            _make_merged_category('WIDE_OVERRIDES', make_single_override('unicode_wide_results', kt)),
            _make_merged_category('SRI_OVERRIDES', make_single_override('sri_results', kt)),
            _make_merged_category('SFZ_OVERRIDES', make_single_override('sfz_results', kt)),
            _make_merged_category('VS16_OVERRIDES', make_single_override('emoji_vs16_results', kt)),
            _make_merged_category('VS15_OVERRIDES', make_single_override('emoji_vs15_results', kt)),
            _make_merged_category('NARROW_OVERRIDES', make_single_override('narrow_results', kt)),
        ])
        yield from fetch_override_grapheme_data(kt)
        yield TermProgramTableRenderDef.new()

    for render_def in get_codegen_definitions():
        print(f'write {render_def.output_filename}: ', flush=True, end='')
        new_filename = render_def.output_filename + '.new'
        with open(new_filename, 'w', encoding='utf-8', newline='\n') as fout:
            for data in render_def.generate():
                fout.write(data)

        os.replace(new_filename, render_def.output_filename)
        print('ok')

    # Update README.rst list_term_programs() example with current terminal names
    if update_readme_term_programs():
        print('updated README.rst: list_term_programs() example')
    else:
        print('README.rst: list_term_programs() example is up-to-date')

    # Remove stale grapheme override files no longer referenced by _registry.py
    cleanup_stale_grapheme_files()


if __name__ == '__main__':
    main(**parse_args())



"""Core tests for wcwidth module."""
# std imports
import importlib.metadata

# 3rd party
import pytest

# local
import wcwidth
from wcwidth._width import _WIDTH_FAST_PATH_MIN_LEN


def test_package_version():
    """wcwidth.__version__ is expected value."""
    # given,
    expected = importlib.metadata.version('wcwidth')

    # exercise,
    result = wcwidth.__version__

    # verify.
    assert result == expected


def test_empty_string():
    """
    Test empty string is OK.

    https://github.com/jquast/wcwidth/issues/24
    """
    phrase = ""
    expect_length_each = 0
    expect_length_phrase = 0

    # exercise,
    length_each = wcwidth.wcwidth(phrase)
    length_phrase = wcwidth.wcswidth(phrase)

    # verify.
    assert length_each == expect_length_each
    assert length_phrase == expect_length_phrase


def basic_string_type():
    """
    This is a python 2-specific test of the basic "string type".

    Such strings cannot contain anything but ascii in python2.
    """
    # given,
    phrase = 'hello\x00world'
    expect_length_each = (1, 1, 1, 1, 1, 0, 1, 1, 1, 1, 1)
    expect_length_phrase = sum(expect_length_each)

    # exercise,
    length_each = tuple(map(wcwidth.wcwidth, phrase))
    length_phrase = wcwidth.wcswidth(phrase)

    # verify.
    assert length_each == expect_length_each
    assert length_phrase == expect_length_phrase


def test_hello_jp():
    """
    Width of Japanese phrase: コンニチハ, セカイ!

    Given a phrase of 5 and 3 Katakana ideographs, joined with 3 English-ASCII punctuation
    characters, totaling 11, this phrase consumes 19 cells of a terminal emulator.
    """
    # given,
    phrase = 'コンニチハ, セカイ!'
    expect_length_each = (2, 2, 2, 2, 2, 1, 1, 2, 2, 2, 1)
    expect_length_phrase = sum(expect_length_each)

    # exercise,
    length_each = tuple(map(wcwidth.wcwidth, phrase))
    length_phrase = wcwidth.wcswidth(phrase)

    # verify.
    assert length_each == expect_length_each
    assert length_phrase == expect_length_phrase


def test_wcswidth_substr():
    """
    Test wcswidth() optional 2nd parameter, ``n``.

    ``n`` determines at which position of the string to stop counting length.
    """
    # given,
    phrase = 'コンニチハ, セカイ!'
    end = 7
    expect_length_each = (2, 2, 2, 2, 2, 1, 1,)
    expect_length_phrase = sum(expect_length_each)

    # exercise,
    length_each = tuple(map(wcwidth.wcwidth, phrase))[:end]
    length_phrase = wcwidth.wcswidth(phrase, end)

    # verify.
    assert length_each == expect_length_each
    assert length_phrase == expect_length_phrase


def test_null_width_0():
    """NULL (0) reports width 0."""
    # given,
    phrase = 'abc\x00def'
    expect_length_each = (1, 1, 1, 0, 1, 1, 1)
    expect_length_phrase = sum(expect_length_each)

    # exercise,
    length_each = tuple(map(wcwidth.wcwidth, phrase))
    length_phrase = wcwidth.wcswidth(phrase, len(phrase))

    # verify.
    assert length_each == expect_length_each
    assert length_phrase == expect_length_phrase


def test_control_c0_width_negative_1():
    """
    How the API reacts to CSI (Control sequence initiate).

    An example of bad fortune, this terminal sequence is a width of 0 on all terminals, but wcwidth
    doesn't parse Control-Sequence-Inducer (CSI) sequences.

    Also the "legacy" posix functions wcwidth and wcswidth return -1 for any string containing the
    C1 control character \x1b (ESC).
    """
    # given,
    phrase = '\x1b[0m'
    expect_length_each = (-1, 1, 1, 1)
    expect_length_phrase = -1

    # exercise,
    length_each = tuple(map(wcwidth.wcwidth, phrase))
    length_phrase = wcwidth.wcswidth(phrase)

    # verify, though this is actually *0* width for a terminal emulator
    assert length_each == expect_length_each
    assert length_phrase == expect_length_phrase


def test_combining_width():
    """Simple test combining reports total width of 4."""
    # given,
    phrase = '--\u05bf--'
    expect_length_each = (1, 1, 0, 1, 1)
    expect_length_phrase = 4

    # exercise,
    length_each = tuple(map(wcwidth.wcwidth, phrase))
    length_phrase = wcwidth.wcswidth(phrase)

    # verify.
    assert length_each == expect_length_each
    assert length_phrase == expect_length_phrase


def test_combining_cafe():
    """Phrase cafe + COMBINING ACUTE ACCENT is café of length 4."""
    phrase = "cafe\u0301"
    expect_length_each = (1, 1, 1, 1, 0)
    expect_length_phrase = 4

    # exercise,
    length_each = tuple(map(wcwidth.wcwidth, phrase))
    length_phrase = wcwidth.wcswidth(phrase)

    # verify.
    assert length_each == expect_length_each
    assert length_phrase == expect_length_phrase


def test_combining_enclosing():
    """CYRILLIC CAPITAL LETTER A + COMBINING CYRILLIC HUNDRED THOUSANDS SIGN is of length 1."""
    phrase = "\u0410\u0488"
    expect_length_each = (1, 0)
    expect_length_phrase = 1

    # exercise,
    length_each = tuple(map(wcwidth.wcwidth, phrase))
    length_phrase = wcwidth.wcswidth(phrase)

    # verify.
    assert length_each == expect_length_each
    assert length_phrase == expect_length_phrase


def test_balinese_script():
    """
    Balinese kapal (ship) is length 3.

    This may be an example that is not yet correctly rendered by any terminal so far, like
    devanagari.
    """
    phrase = ("\u1B13"    # Category 'Lo', EAW 'N' -- BALINESE LETTER KA
              "\u1B28"    # Category 'Lo', EAW 'N' -- BALINESE LETTER PA KAPAL
              "\u1B2E"    # Category 'Lo', EAW 'N' -- BALINESE LETTER LA
              "\u1B44")   # Category 'Mc', EAW 'N' -- BALINESE ADEG ADEG (virama)
    expect_length_each = (1, 1, 1, 0)
    expect_length_phrase = 3

    # exercise,
    length_each = tuple(map(wcwidth.wcwidth, phrase))
    length_phrase = wcwidth.wcswidth(phrase)

    # verify.
    assert length_each == expect_length_each
    assert length_phrase == expect_length_phrase

    # verify width() parse mode also handles Mc correctly
    assert wcwidth.width(phrase) == expect_length_phrase

    # standalone Mc has zero width
    assert wcwidth.wcswidth("\u1B44") == 0
    assert wcwidth.width("\u1B44") == 0


def test_kr_jamo():
    """
    Test basic combining of HANGUL CHOSEONG and JUNGSEONG.

    Example and from Raymond Chen's blog post,
    https://devblogs.microsoft.com/oldnewthing/20201009-00/?p=104351
    """
    # This is an example where both characters are "wide" when displayed alone.
    #
    # But JUNGSEONG (vowel) is designed for combination with a CHOSEONG (consonant).
    #
    # This wcwidth library understands their width only when combination,
    # and not by independent display, like other zero-width characters that may
    # only combine with an appropriate preceding character.
    phrase = (
        "\u1100"  # ᄀ HANGUL CHOSEONG KIYEOK (consonant)
        "\u1161"  # ᅡ HANGUL JUNGSEONG A (vowel)
    )
    expect_length_each = (2, 0)
    expect_length_phrase = 2

    # exercise,
    length_each = tuple(map(wcwidth.wcwidth, phrase))
    length_phrase = wcwidth.wcswidth(phrase)

    # verify.
    assert length_each == expect_length_each
    assert length_phrase == expect_length_phrase


def test_kr_jamo_filler():
    """
    Jamo filler is 0 width.

    Example from https://www.unicode.org/L2/L2006/06310-hangul-decompose9.pdf
    """
    phrase = (
        "\u1100"  # HANGUL CHOSEONG KIYEOK (consonant)
        "\u1160"  # HANGUL JUNGSEONG FILLER (vowel)
    )
    expect_length_each = (2, 0)
    expect_length_phrase = 2

    # exercise,
    length_each = tuple(map(wcwidth.wcwidth, phrase))
    length_phrase = wcwidth.wcswidth(phrase)

    # verify.
    assert length_each == expect_length_each
    assert length_phrase == expect_length_phrase


def test_devanagari_script():
    """
    Attempt to test the measurement width of Devanagari script.

    I believe this 'phrase' should be length 3.

    This is a difficult problem, and this library does not yet get it right,
    because we interpret the unicode data files programmatically, but they do
    not correctly describe how their terminal width is measured.

    There are very few Terminals that do!

    As of 2023,

    - iTerm2: correct length but individual characters are out of order and
              horizaontally misplaced as to be unreadable in its language when
              using 'Noto Sans' font.
    - mlterm: mixed results, it offers several options in the configuration
              dialog, "Xft", "Cario", and "Variable Column Width" have some
              effect, but with neither 'Noto Sans' or 'unifont', it is not
              recognizable as the Devanagari script it is meant to display.

    Previous testing with Devanagari documented at address https://benizi.com/vim/devanagari/

    See also, https://askubuntu.com/questions/8437/is-there-a-good-mono-spaced-font-for-devanagari-script-in-the-terminal
    """
    # This test adapted from https://www.unicode.org/L2/L2023/23107-terminal-suppt.pdf
    # please note that document correctly points out that the final width cannot be determined
    # as a sum of each individual width, as this library currently performs with exception of
    # ZWJ, but I think it incorrectly gestures what a stateless call to wcwidth.wcwidth of
    # each codepoint *should* return.
    phrase = ("\u0915"    # Akhand, Category 'Lo', East Asian Width property 'N' -- DEVANAGARI LETTER KA
              "\u094D"    # Joiner, Category 'Mn', East Asian Width property 'N' -- DEVANAGARI SIGN VIRAMA
              "\u0937"    # Fused, Category 'Lo', East Asian Width property 'N' -- DEVANAGARI LETTER SSA
              "\u093F")   # MatraL, Category 'Mc', East Asian Width property 'N' -- DEVANAGARI VOWEL SIGN I
    # 23107-terminal-suppt.pdf suggests wcwidth.wcwidth should return (2, 0, 0, 1)
    expect_length_each = (1, 0, 1, 0)
    # grapheme cluster capped at 2 cells (ghostty, foot, Windows Terminal)
    expect_length_phrase = 2

    # exercise,
    length_each = tuple(map(wcwidth.wcwidth, phrase))
    length_phrase = wcwidth.wcswidth(phrase)

    # verify.
    assert length_each == expect_length_each
    assert length_phrase == expect_length_phrase
    assert wcwidth.width(phrase) == expect_length_phrase


def test_tamil_script():
    # This test adapted from https://www.unicode.org/L2/L2023/23107-terminal-suppt.pdf
    phrase = ("\u0b95"    # Akhand, Category 'Lo', East Asian Width property 'N' -- TAMIL LETTER KA
              "\u0bcd"    # Joiner, Category 'Mn', East Asian Width property 'N' -- TAMIL SIGN VIRAMA
              "\u0bb7"    # Fused, Category 'Lo', East Asian Width property 'N' -- TAMIL LETTER SSA
              "\u0bcc")   # MatraLR, Category 'Mc', East Asian Width property 'N' -- TAMIL VOWEL SIGN AU
    # 23107-terminal-suppt.pdf suggests wcwidth.wcwidth should return (3, 0, 0, 4)
    expect_length_each = (1, 0, 1, 0)

    # grapheme cluster capped at 2 cells (ghostty, foot, Windows Terminal)
    expect_length_phrase = 2

    # exercise,
    length_each = tuple(map(wcwidth.wcwidth, phrase))
    length_phrase = wcwidth.wcswidth(phrase)

    # verify.
    assert length_each == expect_length_each
    assert length_phrase == expect_length_phrase
    assert wcwidth.width(phrase) == expect_length_phrase


def test_kannada_script():
    # This test adapted from https://www.unicode.org/L2/L2023/23107-terminal-suppt.pdf
    # |ರ್ಝೈ|
    # |123|
    phrase = ("\u0cb0"    # Repha, Category 'Lo', East Asian Width property 'N' -- KANNADA LETTER RA
              "\u0ccd"    # Joiner, Category 'Mn', East Asian Width property 'N' -- KANNADA SIGN VIRAMA
              "\u0c9d"    # Base, Category 'Lo', East Asian Width property 'N' -- KANNADA LETTER JHA
              "\u0cc8")   # MatraUR, Category 'Mc', East Asian Width property 'N' -- KANNADA VOWEL SIGN AI
    # 23107-terminal-suppt.pdf suggests should be (2, 0, 3, 1)
    expect_length_each = (1, 0, 1, 0)
    # grapheme cluster capped at 2 cells (ghostty, foot, Windows Terminal)
    expect_length_phrase = 2

    # exercise,
    length_each = tuple(map(wcwidth.wcwidth, phrase))
    length_phrase = wcwidth.wcswidth(phrase)

    # verify.
    assert length_each == expect_length_each
    assert length_phrase == expect_length_phrase
    assert wcwidth.width(phrase) == expect_length_phrase


def test_kannada_script_2():
    # This test adapted from https://www.unicode.org/L2/L2023/23107-terminal-suppt.pdf
    # |ರ಼್ಚ|
    # |12|
    phrase = ("\u0cb0"    # Base, Category 'Lo', East Asian Width property 'N' -- KANNADA LETTER RA
              "\u0cbc"    # Nukta, Category 'Mn', East Asian Width property 'N' -- KANNADA SIGN NUKTA
              "\u0ccd"    # Joiner, Category 'Lo', East Asian Width property 'N' -- KANNADA SIGN VIRAMA
              "\u0c9a")   # Subjoin, Category 'Mc', East Asian Width property 'N' -- KANNADA LETTER CA
    # 23107-terminal-suppt.pdf suggests wcwidth.wcwidth should return (2, 0, 0, 1)
    expect_length_each = (1, 0, 0, 1)
    # virama conjunct collapses RA(+Nukta)+virama+CA into one cell
    expect_length_phrase = 2

    # exercise,
    length_each = tuple(map(wcwidth.wcwidth, phrase))
    length_phrase = wcwidth.wcswidth(phrase)

    # verify.
    assert length_each == expect_length_each
    assert length_phrase == expect_length_phrase
    assert wcwidth.width(phrase) == expect_length_phrase


def test_bengali_nukta_mc():
    # Mc following Mn (Nukta) is still counted as spacing mark.
    #
    # Discovered via UDHR Bengali text where wrap() produced lines exceeding the requested width.
    # The root cause was that width() only recognized a Spacing Combining Mark (Mc) when it was
    # *immediately* adjacent to the base character (index == last_base + 1).
    #
    # In Bengali, a Nukta (U+09BC, category Mn) commonly sits between the consonant and the vowel
    # sign, so the Mc vowel sign was skipped and measured as zero instead of one.
    #
    # The nukta between consonant and vowel sign does not break the combining sequence, so the Mc
    # must still be counted.
    phrase = "\u09AF\u09BC\u09C7"
    assert wcwidth.wcwidth("\u09C7") == 0
    assert wcwidth.wcswidth(phrase) == 2
    assert wcwidth.width(phrase) == 2


@pytest.mark.parametrize("repeat", [1, _WIDTH_FAST_PATH_MIN_LEN])
def test_mc_width_consistency(repeat):
    """Check width() to wcswidth() consistency."""
    # repeat value 'WIDTH_FAST_PATH_MIN_LEN' ensures both "fast" and "slow" paths are taken
    phrases = [
        "\u0915\u094D\u0937\u093F",
        "\u0b95\u0bcd\u0bb7\u0bcc",
        "\u0cb0\u0ccd\u0c9d\u0cc8",
        "\u0cb0\u0cbc\u0ccd\u0c9a",
        "\u09AF\u09BC\u09C7",
        "\u09B9\u09AF\u09BC\u09C7\u099B\u09C7",
        "\u0915\u09BE\u0999\u09CD\u0996\u09BE",
    ]
    # Virama conjunct collapsing is context-sensitive across grapheme
    # boundaries (virama ends one grapheme, consonant starts the next),
    # so per-grapheme width sums may exceed wcswidth/width totals for
    # phrases containing conjuncts.
    no_conjunct_phrases = [
        "\u09AF\u09BC\u09C7",
    ]
    for phrase in phrases:
        text = phrase * repeat
        assert wcwidth.width(text) == wcwidth.wcswidth(text)
    for phrase in no_conjunct_phrases:
        text = phrase * repeat
        grapheme_sum = sum(wcwidth.width(g) for g in wcwidth.iter_graphemes(text))
        assert wcwidth.width(text) == grapheme_sum


@pytest.mark.parametrize("phrase,expected", [
    ("\u0999\u09CD\u0997\u09C7", 2),
    ("\u0915\u094D\u0924\u093F", 2),
    ("\u0915\u094D\u0930\u093F", 2),
    ("\u0A95\u0ACD\u0A95\u0ACB", 2),
    ("\u0938\u094D\u0924\u094D\u0930", 2),
    ("\u0938\u094D\u0924", 2),
    ("\u0915\u094D\u0020", 2),
    ("\u09A4\u09CD\u200D\u09AA", 2),
    ("\u0915\u094D\u200D\u0924", 2),
    ("\u0D15\u0D4D\u0D15\u0D41\u0D02", 2),
    ("\u0915\u094D\u0924\u0941\u0902", 2),
])
def test_virama_conjunct(phrase, expected):
    assert wcwidth.wcswidth(phrase) == expected
    assert wcwidth.width(phrase) == expected


@pytest.mark.parametrize("phrase,expected", [
    ("\u0995\u09CD\u09A4\u09BF", 2),       # Bengali C+V+C+Mc: ক্তি (capped at 2)
    ("\u0915\u094D\u0924\u093F", 2),       # Devanagari C+V+C+Mc: क्ति (capped at 2)
    ("\u0995\u09CD\u09A4", 2),             # C+V+C (no Mc), unchanged
    ("\u0995\u09BF", 2),                   # C+Mc (no virama), unchanged
])
def test_virama_conjunct_mc_vowel(phrase, expected):
    """Mc combines into base; cluster capped at 2."""
    assert wcwidth.wcswidth(phrase) == expected
    assert wcwidth.width(phrase) == expected
    assert wcwidth.wcswidth(phrase) == expected
    assert wcwidth.width(phrase) == expected


@pytest.mark.parametrize("phrase,expected", [
    ("\uA9A0\uA9C0\uA9B1\uA9C0\uA9AE", 2),  # Javanese C+V+C+V+C: TA+PANGKON+SA+PANGKON+WA
    ("\uA9A0\uA9C0\uA9B1", 2),               # Javanese C+V+C: TA+PANGKON+SA
    ("\u1B04\u1B44\u1B05", 2),               # Balinese C+V+C: A+ADEG ADEG+I
    ("\U000111C0\U000111C0", 0),             # Sharada virama alone (zero-width)
])
def test_virama_mc_category_overlap(phrase, expected):
    """Virama codepoints in Mc category check ISC before Mc."""
    assert wcwidth.wcswidth(phrase) == expected
    assert wcwidth.width(phrase) == expected


@pytest.mark.parametrize("phrase,expected", [
    ("\u1000\u1039\u1000", 2),             # Burmese KA+VIRAMA+KA
    ("\u1000\u1039\u1000\u1039\u1002", 2),  # Burmese KA+V+KA+V+GA (capped)
    ("\u1000\u1039\u200D\u1000", 2),       # Burmese KA+V+ZWJ+KA
    ("\u1782\u17D2\u1782\u17C1", 2),       # Khmer KO+COENG+KO+VOWEL_E (capped)
    ("\u1780\u17D2\u1780", 2),             # Khmer KA+COENG+KA
])
def test_virama_conjunct_invisible_stacker(phrase, expected):
    assert wcwidth.wcswidth(phrase) == expected
    assert wcwidth.width(phrase) == expected


def test_zwj_at_end_of_string():
    """ZWJ at end of string (not after virama) is consumed with zero width."""
    assert wcwidth.wcswidth('a\u200D') == 1


def test_wcswidth_n_exceeds_length():
    """Verify wcswidth() with n > len(string) does not raise IndexError."""
    assert wcwidth.wcswidth('hello', n=999) == 5
    assert wcwidth.wcswidth('\u30B3\u30F3', n=999) == 4


def test_soft_hyphen():
    # Test SOFT HYPHEN, category 'Cf' usually are zero-width, but most
    # implementations agree to draw it was '1' cell, visually
    # indistinguishable from a space, ' ' in Konsole, for example.
    assert wcwidth.wcwidth(chr(0x000ad)) == 1


PREPENDED_CONCATENATION_MARKS = [
    (0x0600, 'ARABIC NUMBER SIGN'),
    (0x0601, 'ARABIC SIGN SANAH'),
    (0x0602, 'ARABIC FOOTNOTE MARKER'),
    (0x0603, 'ARABIC SIGN SAFHA'),
    (0x0604, 'ARABIC SIGN SAMVAT'),
    (0x0605, 'ARABIC NUMBER MARK ABOVE'),
    (0x06DD, 'ARABIC END OF AYAH'),
    (0x070F, 'SYRIAC ABBREVIATION MARK'),
    (0x0890, 'ARABIC POUND MARK ABOVE'),
    (0x0891, 'ARABIC PIASTRE MARK ABOVE'),
    (0x08E2, 'ARABIC DISPUTED END OF AYAH'),
    (0x110BD, 'KAITHI NUMBER SIGN'),
    (0x110CD, 'KAITHI NUMBER SIGN ABOVE'),
]


@pytest.mark.parametrize('codepoint,name', PREPENDED_CONCATENATION_MARKS)
def test_prepended_concatenation_mark_width(codepoint, name):
    """Prepended Concatenation Marks have width 1, not 0."""
    # https://github.com/jquast/wcwidth/issues/119
    assert wcwidth.wcwidth(chr(codepoint)) == 1


def test_legacy_module():
    """Verify legacy ``wcwidth.wcwidth`` module's public items are importable."""
    # pylint: disable=import-outside-toplevel
    # Save and restore wcwidth.wcwidth because importing the submodule
    # rebinds the package attribute from the function to the module.
    _wcwidth_func = wcwidth.wcwidth
    _legacy = __import__('wcwidth.wcwidth', fromlist=['wcwidth'])
    wcwidth.wcwidth = _wcwidth_func

    for name in _legacy.__all__:
        attr = getattr(_legacy, name)
        assert attr is not None, f"wcwidth.wcwidth.{name} is None"

    # Verify that individual imports from the legacy path also work,
    # e.g. 'from wcwidth.wcwidth import wcswidth'
    for name in _legacy.__all__:
        obj = getattr(_legacy, name)
        assert obj is not None, f"could not import {name} from wcwidth.wcwidth"



"""Unicode version level tests for wcwidth."""

# local
import wcwidth


def test_list_versions_single():
    """list_versions returns only the latest version."""
    versions = wcwidth.list_versions()
    assert len(versions) == 1
    assert versions[0] == "17.0.0"


def test_latest():
    """wcwidth._wcmatch_version('latest') returns the latest version."""
    expected = wcwidth.list_versions()[-1]
    result = wcwidth._wcmatch_version('latest')
    assert result == expected


def test_auto():
    """wcwidth._wcmatch_version('auto') returns the latest version."""
    expected = wcwidth.list_versions()[-1]
    result = wcwidth._wcmatch_version('auto')
    assert result == expected


def test_wcmatch_version_always_latest():
    """_wcmatch_version always returns latest regardless of input."""
    latest = wcwidth.list_versions()[-1]
    assert wcwidth._wcmatch_version('4.1.0') == latest
    assert wcwidth._wcmatch_version('9.0.0') == latest
    assert wcwidth._wcmatch_version('auto') == latest
    assert wcwidth._wcmatch_version('latest') == latest
    assert wcwidth._wcmatch_version('invalid') == latest
    assert wcwidth._wcmatch_version('999.0.0') == latest
    assert wcwidth._wcmatch_version('x.y.z') == latest
    assert wcwidth._wcmatch_version('8') == latest
    assert wcwidth._wcmatch_version('5.0.5') == latest


def test_env_var_ignored(monkeypatch):
    """UNICODE_VERSION environment variable is ignored."""
    monkeypatch.setenv('UNICODE_VERSION', '4.1.0')
    wcwidth._wcmatch_version.cache_clear()
    assert wcwidth._wcmatch_version('auto') == wcwidth.list_versions()[-1]


def test_unicode_version_param_ignored():
    """unicode_version parameter is ignored in wcwidth/wcswidth."""
    # wcwidth should work the same regardless of unicode_version value
    assert wcwidth.wcwidth('A', unicode_version='4.1.0') == 1
    assert wcwidth.wcwidth('A', unicode_version='latest') == 1
    assert wcwidth.wcwidth('A', unicode_version='invalid') == 1

    # wcswidth should work the same regardless of unicode_version value
    assert wcwidth.wcswidth('hello', unicode_version='4.1.0') == 5
    assert wcwidth.wcswidth('hello', unicode_version='latest') == 5
    assert wcwidth.wcswidth('hello', unicode_version='invalid') == 5



"""Tests for sequence-aware text wrapping functions."""

# std imports
import sys
import platform
import textwrap

# 3rd party
import pytest

# local
from wcwidth import iter_sequences
from wcwidth.textwrap import SequenceTextWrapper, wrap


@pytest.fixture(autouse=True)
def mock_hyperlink_ids(monkeypatch):
    """Mock secrets.token_hex to return predictable IDs for testing."""
    counter = 0

    def fake_token_hex(n):
        nonlocal counter
        counter += 1
        return f'{counter:0{n * 2}x}'

    monkeypatch.setattr('secrets.token_hex', fake_token_hex)


SGR_RED = '\x1b[31m'
SGR_BOLD = '\x1b[1m'
SGR_RESET = '\x1b[0m'
ATTRS = ('\x1b[31m', '\x1b[34m', '\x1b[4m', '\x1b[7m', '\x1b[41m', '\x1b[37m', '\x1b[107m')

OSC_HYPERLINK = '\x1b]8;;https://example.com\x07link\x1b]8;;\x07'
CSI_CURSOR = '\x1b[5C'
CTRL_BEL = '\x07'

ZWJ = '\u200d'
WOMAN = '\U0001F469'
GIRL = '\U0001F467'
FAMILY_ZWJ = f'{WOMAN}{ZWJ}{WOMAN}{ZWJ}{GIRL}'
SMILEY_VS16 = '\u263a\ufe0f'
ZWJ_FAMILY = '\U0001F469\u200D\U0001F469\u200D\U0001F467\u200D\U0001F466'
CAFE_COMBINING = 'cafe\u0301'
HANGUL_GA = '\u1100\u1161'


def _strip(text):
    return ''.join(seg for seg, is_seq in iter_sequences(text) if not is_seq)


def _adjust_stdlib_result(expected, kwargs):
    """
    Adjust stdlib textwrap result for known bugs in older Python versions.

    CPython #140627: Older versions leave trailing whitespace and preceding all-whitespace lines
    when drop_whitespace=True. Fixed in 3.13.11+, 3.14.2+, and 3.15+. We always strip to normalize
    across versions.
    """
    if not expected:
        return expected
    if kwargs.get('drop_whitespace', True):
        # Strip trailing whitespace from each line (old Python bug)
        expected = [line.rstrip() for line in expected]
        # Remove leading all-whitespace lines (old Python bug)
        if expected and not expected[0].strip():
            expected = expected[1:]
            if expected and kwargs.get('subsequent_indent'):
                expected[0] = expected[0][len(kwargs['subsequent_indent']):]
    return expected


def _colorize(text):
    return ''.join(
        ATTRS[idx % len(ATTRS)] + char + SGR_RESET if char not in ' -\t' else char
        for idx, char in enumerate(text)
    )


@pytest.mark.parametrize('text,w,expected', [
    ('', 10, []),
    ('   ', 10, []),
    ('\u5973', 0, ['\u5973']),
    ('\u5973', 1, ['\u5973']),
    (ZWJ_FAMILY, 1, [ZWJ_FAMILY]),
    (HANGUL_GA, 1, [HANGUL_GA]),
])
def test_wrap_edge_cases(text, w, expected):
    assert wrap(text, w) == expected


def test_wrap_initial_indent():
    assert wrap('hello world', 10, initial_indent='> ') == ['> hello', 'world']


@pytest.mark.parametrize('text,w,break_long,expected', [
    ('abcdefghij', 3, True, ['abc', 'def', 'ghi', 'j']),
    ('abcdefghij', 3, False, ['abcdefghij']),
])
def test_wrap_long_words(text, w, break_long, expected):
    assert wrap(text, w, break_long_words=break_long) == expected


@pytest.mark.parametrize('text,w,break_hyphens,propagate,expected', [
    ('a-b-c-d', 3, True, True, ['a-', 'b-', 'c-d']),
    ('a-b-c-d', 3, False, True, ['a-b', '-c-', 'd']),
    ('---', 2, True, True, ['--', '-']),
    ('a---b', 2, True, True, ['a-', '--', 'b']),
    ('a-\x1b[31mb', 2, True, True, ['a-\x1b[31m\x1b[0m', '\x1b[31mb\x1b[0m']),
    ('a-\x1b[31mb', 2, True, False, ['a-\x1b[31m', 'b']),
])
def test_wrap_hyphen_long_words(text, w, break_hyphens, propagate, expected):
    assert wrap(text, w, break_on_hyphens=break_hyphens, propagate_sgr=propagate) == expected


# Comprehensive stdlib compatibility
TEXTWRAP_KWARGS = [
    {'break_long_words': False, 'drop_whitespace': False},
    {'break_long_words': False, 'drop_whitespace': True},
    {'break_long_words': True, 'drop_whitespace': False},
    {'break_long_words': True, 'drop_whitespace': True},
    {'break_long_words': True, 'drop_whitespace': False, 'subsequent_indent': ' '},
    {'break_long_words': True, 'drop_whitespace': True, 'subsequent_indent': ' '},
    {'break_long_words': True, 'drop_whitespace': True, 'break_on_hyphens': True},
    {'break_long_words': True, 'drop_whitespace': True, 'break_on_hyphens': False},
    {'break_long_words': True, 'drop_whitespace': False,
     'subsequent_indent': '', 'max_lines': 4, 'placeholder': '~'},
    {'break_long_words': True, 'drop_whitespace': True,
     'max_lines': 3, 'placeholder': '...'},
    {'break_long_words': True, 'drop_whitespace': True,
     'max_lines': 1, 'placeholder': '...'},
    {'expand_tabs': False, 'break_long_words': True, 'drop_whitespace': True},
    {'replace_whitespace': False, 'break_long_words': True,
     'drop_whitespace': True},
    {'fix_sentence_endings': True, 'break_long_words': True,
     'drop_whitespace': True},
]


@pytest.mark.parametrize('kwargs', TEXTWRAP_KWARGS)
@pytest.mark.parametrize('width', [3, 8, 20, 33])
def test_wrap_matches_stdlib(kwargs, width):
    pgraph = ' Z! a bc defghij klmnopqrstuvw<<>>xyz012345678900 ' * 2
    pgraph_colored = _colorize(pgraph)
    expected = _adjust_stdlib_result(
        textwrap.wrap(pgraph, width=width, **kwargs), kwargs
    )
    wrapper = SequenceTextWrapper(width=width, **kwargs)
    assert wrapper.wrap(pgraph) == expected
    # For colored text, strip sequences
    colored_result = [_strip(line) for line in wrapper.wrap(pgraph_colored)]
    if kwargs.get('drop_whitespace'):
        # normalize trailing whitespace, rstrip when drop_whitespace is True
        # matches CPython #140627 fix
        colored_result = [line.rstrip() for line in colored_result]
    assert colored_result == expected


@pytest.mark.parametrize('kwargs', TEXTWRAP_KWARGS)
@pytest.mark.parametrize('width', [8, 20, 37])
@pytest.mark.parametrize('tabsize', [4, 5, 8])
def test_wrap_tabsize_matches_stdlib(kwargs, width, tabsize):
    tabsize = min(tabsize, width)
    pgraph = ' Z! a bc\t defghij\t kl mnopqrs\ttuvw<<>>xyz012345678900 ' * 2
    expected = _adjust_stdlib_result(
        textwrap.wrap(pgraph, width=width, tabsize=tabsize, **kwargs), kwargs
    )
    wrapper = SequenceTextWrapper(width=width, tabsize=tabsize, **kwargs)
    assert wrapper.wrap(pgraph) == expected


def test_wrap_multiline_matches_stdlib():
    given = '\n' + 32 * 'A' + '\n' + 32 * 'B' + '\n' + 32 * 'C' + '\n\n'
    assert wrap(given, 30) == textwrap.wrap(given, 30)


@pytest.mark.parametrize('text,w,expected', [
    # CJK (2 cells each)
    ('\u4e2d\u6587\u5b57\u7b26', 4, ['\u4e2d\u6587', '\u5b57\u7b26']),
    ('\u4e2d\u6587\u5b57', 5, ['\u4e2d\u6587', '\u5b57']),
    # Combining characters
    (CAFE_COMBINING + '-latte', 4, ['cafe\u0301', '-lat', 'te']),
    # Emoji (ZWJ, VS16)
    (f'{FAMILY_ZWJ} ab', 4, [FAMILY_ZWJ, 'ab']),
    (f'{SMILEY_VS16} ab', 3, [SMILEY_VS16, 'ab']),
    ('\U0001F469\U0001F467\U0001F466', 4, ['\U0001F469\U0001F467', '\U0001F466']),
])
def test_wrap_unicode(benchmark, text, w, expected):
    kwargs = {'break_on_hyphens': False} if '-' in text else {}
    result = benchmark(wrap, text, w, **kwargs)
    assert result == expected


@pytest.mark.parametrize('text,w,expected', [
    # SGR sequences propagated across lines
    (f'{SGR_RED}red{SGR_RESET} blue', 4, [f'{SGR_RED}red{SGR_RESET}', 'blue']),
    # SGR at end of line propagates to next line
    (f'hello{SGR_RED} world', 6, [f'hello{SGR_RED}{SGR_RESET}', f'{SGR_RED}world{SGR_RESET}']),
    # Empty/adjacent sequences - sequences kept (no visible content)
    (f'{SGR_RED}{SGR_RESET}', 10, [f'{SGR_RED}{SGR_RESET}']),
    # Reset clears style, sequences kept with second line
    (f'hello {SGR_RED}{SGR_RESET}world', 6, ['hello', f'{SGR_RED}{SGR_RESET}world']),
    # OSC hyperlinks (with space separator) - not SGR, so preserved as-is
    (f'{OSC_HYPERLINK} text', 5, [OSC_HYPERLINK, 'text']),
    # CSI cursor sequences - not SGR, so preserved as-is
    (f'{CSI_CURSOR}text here', 10, [f'{CSI_CURSOR}text', 'here']),
    # Control characters
    (f'{CTRL_BEL}alert text', 6, [f'{CTRL_BEL}alert', 'text']),
    # Sequences in long word breaking - red starts after 'x', continues across lines
    ('x\x1b[31mabcdefghij\x1b[0m', 3,
     ['x\x1b[31mab\x1b[0m', '\x1b[31mcde\x1b[0m', '\x1b[31mfgh\x1b[0m', '\x1b[31mij\x1b[0m']),
    # Fs sequence (ESC d) - zero-width, stays with preceding text
    ('abc\x1bdefghij', 3, ['abc\x1bd', 'efg', 'hij']),
])
def test_wrap_sequences(benchmark, text, w, expected):
    """Escape sequence preservation (with propagate_sgr=True default)"""
    assert benchmark(wrap, text, w) == expected


@pytest.mark.parametrize('text,w,expected', [
    (f'hello{SGR_RED} world', 6, [f'hello{SGR_RED}', 'world']),
    ('x\x1b[31mabcdefghij\x1b[0m', 3, ['x\x1b[31mab', 'cde', 'fgh', 'ij\x1b[0m']),
]
)
def test_wrap_sequences_no_propagate(text, w, expected):
    result = wrap(text, w, propagate_sgr=False)
    assert result == expected


@pytest.mark.parametrize('text,w,expected', [
    (f'{SGR_RED}\u4e2d\u6587{SGR_RESET} ab', 5, [f'{SGR_RED}\u4e2d\u6587{SGR_RESET}', 'ab']),
    (f'{SGR_RED}{FAMILY_ZWJ}{SGR_RESET} ab', 4, [f'{SGR_RED}{FAMILY_ZWJ}{SGR_RESET}', 'ab']),
    (f'{SGR_BOLD}\u4e2d{SGR_RESET}y z', 4, [f'{SGR_BOLD}\u4e2d{SGR_RESET}y', 'z']),
])
def test_wrap_mixed(benchmark, text, w, expected):
    """Test mixed sequences + unicode."""
    result = benchmark(wrap, text, w)
    assert result == expected


@pytest.mark.parametrize('text,w,tabsize,expected', [
    # CJK (2 cells) + tab: tabsize=4, '\u4e2d' is 2 cols, tab expands to col 4
    ('\u4e2d\ta b', 6, 4, ['\u4e2d   a', 'b']),
    # CJK + tab with tabsize=8: '\u4e2d' is 2 cols, tab expands to col 8
    ('\u4e2d\ta b', 10, 8, ['\u4e2d       a', 'b']),
    # Emoji + tab (emoji width=2): similar column alignment
    (f'{SMILEY_VS16}\ta b', 6, 4, [f'{SMILEY_VS16}  a', 'b']),
    # Multiple CJK + tab: 4 cols, tab to 4 adds 0, but expand_tabs adds min 1
    ('\u4e2d\u6587\ta', 8, 4, ['\u4e2d\u6587  a']),
    # ASCII + tab + CJK: 'a' is 1 col, tab to 4 (3 spaces), CJK is 2 cols
    ('a\t\u4e2d b', 8, 4, ['a   \u4e2d b']),
])
@pytest.mark.skipif(
    platform.python_implementation() == 'PyPy' and sys.version_info < (3, 9),
    reason='PyPy 3.8 str.expandtabs() counts UTF-8 bytes instead of characters'
)
def test_wrap_tabsize_wide_chars(text, w, tabsize, expected):
    """Verify tabsize respects wide character column positions."""
    assert wrap(text, w, tabsize=tabsize) == expected


OSC_START_ST = '\x1b]8;;http://example.com\x1b\\'
OSC_END_ST = '\x1b]8;;\x1b\\'
OSC_START_BEL = '\x1b]8;;http://example.com\x07'
OSC_END_BEL = '\x1b]8;;\x07'


@pytest.mark.parametrize('text,w,expected', [
    (   # standard, ST-variant,
        f'{OSC_START_ST}link{OSC_END_ST}more',
        5,
        [f'{OSC_START_ST}link{OSC_END_ST}', 'more'],
    ),
    (   # BEL-variant,
        f'{OSC_START_BEL}link{OSC_END_BEL}more',
        5,
        [f'{OSC_START_BEL}link{OSC_END_BEL}', 'more'],
    ),
    (   # hyperlink breaks after word, 'prefix',
        f'prefix{OSC_START_ST}link{OSC_END_ST}',
        6,
        ['prefix', f'{OSC_START_ST}link{OSC_END_ST}'],
    ),
    (
        f'prefix{OSC_START_BEL}link{OSC_END_BEL}',
        6,
        ['prefix', f'{OSC_START_BEL}link{OSC_END_BEL}'],
    ),
    (   # hyperlink breaks before following, 'suffix',
        f'prefix{OSC_START_ST}link{OSC_END_ST}suffix',
        6,
        ['prefix', f'{OSC_START_ST}link{OSC_END_ST}', 'suffix'],
    ),
    (
        f'prefix{OSC_START_BEL}link{OSC_END_BEL}suffix',
        6,
        ['prefix', f'{OSC_START_BEL}link{OSC_END_BEL}', 'suffix'],
    ),
    (   # hyperlink *surrounded* by SGR attributes
        f'foo {SGR_RED}{OSC_START_ST}link{OSC_END_ST}{SGR_RESET} bar',
        6,
        ['foo', f'{SGR_RED}{OSC_START_ST}link{OSC_END_ST}{SGR_RESET}', 'bar'],
    ),
    (
        f'foo {SGR_RED}{OSC_START_BEL}link{OSC_END_BEL}{SGR_RESET} bar',
        6,
        ['foo', f'{SGR_RED}{OSC_START_BEL}link{OSC_END_BEL}{SGR_RESET}', 'bar'],
    ),
    (   # hyperlink *containing* SGR attributes
        f'foo {OSC_START_ST}{SGR_RED}link{SGR_RESET}{OSC_END_ST} bar',
        6,
        ['foo', f'{OSC_START_ST}{SGR_RED}link{SGR_RESET}{OSC_END_ST}', 'bar'],
    ),
    (
        f'foo {OSC_START_BEL}{SGR_RED}link{SGR_RESET}{OSC_END_BEL} bar',
        6,
        ['foo', f'{OSC_START_BEL}{SGR_RED}link{SGR_RESET}{OSC_END_BEL}', 'bar'],
    ),
    (   # hyperlink with internal space - breaks with id continuation (ST)
        f'Go {OSC_START_ST}Click here{OSC_END_ST} now',
        5,
        [
            'Go',
            '\x1b]8;id=00000001;http://example.com\x1b\\Click\x1b]8;;\x1b\\',
            '\x1b]8;id=00000001;http://example.com\x1b\\here\x1b]8;;\x1b\\',
            'now',
        ],
    ),
    (   # hyperlink with internal space - breaks with id continuation (BEL)
        f'Go {OSC_START_BEL}Click here{OSC_END_BEL} now',
        5,
        [
            'Go',
            '\x1b]8;id=00000001;http://example.com\x07Click\x1b]8;;\x07',
            '\x1b]8;id=00000001;http://example.com\x07here\x1b]8;;\x07',
            'now',
        ],
    ),
    (   # hyperlink with existing id= parameter is preserved
        '\x1b]8;id=my-link;http://example.com\x1b\\Click here\x1b]8;;\x1b\\',
        6,
        [
            '\x1b]8;id=my-link;http://example.com\x1b\\Click\x1b]8;;\x1b\\',
            '\x1b]8;id=my-link;http://example.com\x1b\\here\x1b]8;;\x1b\\',
        ],
    ),
    (   # hyperlink spanning 3+ lines
        f'{OSC_START_ST}one two three{OSC_END_ST}',
        5,
        [
            '\x1b]8;id=00000001;http://example.com\x1b\\one\x1b]8;;\x1b\\',
            '\x1b]8;id=00000001;http://example.com\x1b\\two\x1b]8;;\x1b\\',
            '\x1b]8;id=00000001;http://example.com\x1b\\three\x1b]8;;\x1b\\',
        ],
    ),
    (   # multiple hyperlinks in same text
        f'{OSC_START_ST}ab cd{OSC_END_ST} {OSC_START_BEL}ef gh{OSC_END_BEL}',
        4,
        [
            '\x1b]8;id=00000001;http://example.com\x1b\\ab\x1b]8;;\x1b\\',
            '\x1b]8;id=00000001;http://example.com\x1b\\cd\x1b]8;;\x1b\\',
            '\x1b]8;id=00000002;http://example.com\x07ef\x1b]8;;\x07',
            '\x1b]8;id=00000002;http://example.com\x07gh\x1b]8;;\x07',
        ],
    ),
    (   # long word inside hyperlink forces character-level breaking
        f'{OSC_START_ST}abcdefgh{OSC_END_ST}',
        3,
        [
            '\x1b]8;id=00000001;http://example.com\x1b\\abc\x1b]8;;\x1b\\',
            '\x1b]8;id=00000001;http://example.com\x1b\\def\x1b]8;;\x1b\\',
            '\x1b]8;id=00000001;http://example.com\x1b\\gh\x1b]8;;\x1b\\',
        ],
    ),
    (   # params with other keys but no id - id is prepended, other params preserved
        '\x1b]8;foo=bar;http://example.com\x1b\\Click here\x1b]8;;\x1b\\',
        6,
        [
            '\x1b]8;id=00000001:foo=bar;http://example.com\x1b\\Click\x1b]8;;\x1b\\',
            '\x1b]8;id=00000001:foo=bar;http://example.com\x1b\\here\x1b]8;;\x1b\\',
        ],
    ),
    (   # id not at start of params (junk:id=given) - full params preserved
        '\x1b]8;foo=bar:id=mylink;http://example.com\x1b\\Click here\x1b]8;;\x1b\\',
        6,
        [
            '\x1b]8;foo=bar:id=mylink;http://example.com\x1b\\Click\x1b]8;;\x1b\\',
            '\x1b]8;foo=bar:id=mylink;http://example.com\x1b\\here\x1b]8;;\x1b\\',
        ],
    ),
    (   # wide grapheme after OSC 8 open at width 1 (must not hang)
        '\x1b]8;;u\x07😀',
        1,
        ['\x1b]8;id=00000001;u\x07😀\x1b]8;;\x07'],
    ),
    (   # CJK wide char after OSC 8 open at width 1 (must not hang)
        '\x1b]8;;u\x07あ',
        1,
        ['\x1b]8;id=00000001;u\x07あ\x1b]8;;\x07'],
    ),])
def test_wrap_hyperlink_word_boundary(text, w, expected):
    """OSC hyperlink sequences should act as word boundaries."""
    result = wrap(text, w)
    assert result == expected


@pytest.mark.parametrize('text,kwargs', [
    ('The quick brown fox jumps over the lazy dog',
     {'width': 10, 'max_lines': 3, 'placeholder': '...'}),
    ('1234567890 1234567890 extra',
     {'width': 10, 'max_lines': 2, 'placeholder': '...'}),
    ('1234567890 1234567890',
     {'width': 10, 'max_lines': 1, 'placeholder': '...'}),
    ('short 1234567890 extra',
     {'width': 10, 'max_lines': 2, 'placeholder': '...'}),
    ('hello world',
     {'width': 10, 'max_lines': 5, 'placeholder': '...'}),
    ('hello world foo bar baz',
     {'width': 8, 'max_lines': 2, 'placeholder': ' [...]'}),
    ('a bb ccc',
     {'width': 3, 'max_lines': 2, 'placeholder': '.'}),
    ('a bb ccc dddd',
     {'width': 4, 'max_lines': 3, 'placeholder': '~'}),
    (' a ', {'width': 1, 'max_lines': 2, 'drop_whitespace': False, 'placeholder': '~'}),
    ('hello world', {'width': 20, 'max_lines': 5}),
    ('ab cd', {'width': 5, 'max_lines': 1}),
    ('aaa bbb ccc', {'width': 3, 'max_lines': 2, 'placeholder': '...'}),
    ('hello world foo bar',
     {'width': 10, 'subsequent_indent': '  ', 'max_lines': 2, 'placeholder': '...'}),
    ('hello world foo bar',
     {'width': 10, 'initial_indent': '> ', 'max_lines': 2, 'placeholder': '...'}),
])
def test_wrap_max_lines_matches_stdlib(text, kwargs):
    expected = _adjust_stdlib_result(textwrap.wrap(text, **kwargs), kwargs)
    assert wrap(text, **kwargs) == expected


def test_wrap_placeholder_too_large():
    with pytest.raises(ValueError, match="placeholder too large"):
        wrap('hello', width=3, max_lines=1, placeholder='.....')
    with pytest.raises(ValueError):
        textwrap.wrap('fox', width=1, max_lines=3, placeholder='...')


@pytest.mark.parametrize('text,w,ml,ph,expected', [
    (f'{SGR_RED}hello world foo bar{SGR_RESET}',
     8, 2, '...', [f'{SGR_RED}hello{SGR_RESET}', f'{SGR_RED}world...{SGR_RESET}']),
    (f'{SGR_RED}hello{SGR_RESET} world foo',
     8, 2, '...', [f'{SGR_RED}hello{SGR_RESET}', 'world...']),
    (f'{SGR_RED}hello{SGR_RESET} world',
     6, 1, '.', [f'{SGR_RED}hello{SGR_RESET}.']),
    ('\u4e2d\u6587 \u5b57\u7b26 hello', 5, 1, '~', ['\u4e2d\u6587~']),
    ('\u4e2d\u6587 \u5b57\u7b26 hello world', 5, 2, '~', ['\u4e2d\u6587', '\u5b57\u7b26~']),
    ('\u4e2d\u6587\u5b57\u7b26 hello', 12, 1, '...', ['\u4e2d\u6587\u5b57\u7b26...']),
])
def test_wrap_max_lines_sequences(text, w, ml, ph, expected):
    assert wrap(text, w, max_lines=ml, placeholder=ph) == expected


def test_wrap_max_lines_hyperlink_closed():
    """Truncation inside a hyperlink closes it before the placeholder."""
    text = f'{OSC_START_ST}Click here please{OSC_END_ST}'
    result = wrap(text, 10, max_lines=1, placeholder='...')
    assert len(result) == 1
    assert result[0].endswith(f'{OSC_END_ST}...')
    assert OSC_START_ST in result[0]


def test_wrap_max_lines_hyperlink_close_on_prev_line():
    """Fallback to previous line preserves hyperlink close sequence."""
    text = f'{OSC_START_ST}ab{OSC_END_ST} cccccccccc ddddd'
    result = wrap(text, 10, max_lines=2, placeholder='...')
    assert result == [f'{OSC_START_ST}ab{OSC_END_ST}...']


@pytest.mark.parametrize('text,kwargs', [
    ('hello\tworld', {'width': 20, 'expand_tabs': False, 'replace_whitespace': False}),
    ('hello\tworld foo\tbar baz', {'width': 12, 'expand_tabs': False, 'tabsize': 8}),
    ('hello\nworld', {'width': 20, 'replace_whitespace': False}),
    ('a\t b\n c', {'width': 20, 'replace_whitespace': False}),
    ('Hello world. This is a test.  More text.', {'width': 20, 'fix_sentence_endings': True}),
    ('Dr. Smith went to Washington. He left.', {'width': 20, 'fix_sentence_endings': True}),
])
def test_wrap_stdlib_params(text, kwargs):
    assert wrap(text, **kwargs) == textwrap.wrap(text, **kwargs)


def test_wrap_expand_tabs_false_with_sequences():
    text = f'{SGR_RED}a\tb{SGR_RESET}'
    result = wrap(text, 20, expand_tabs=False, replace_whitespace=False)
    assert '\t' in _strip(result[0])


def test_wrap_replace_whitespace_false_newlines_zero_width():
    """Newlines have zero display width, so more text fits per line than stdlib."""
    assert wrap('hello\nworld foo\nbar', 10, replace_whitespace=False) == [
        'hello\nworld', 'foo\nbar']


def test_wrap_bare_esc():
    """Bare ESC not part of a recognized sequence is treated as zero-width."""
    assert wrap('ab\x1bcd ef', 5) == ['ab\x1bcd', 'ef']
    assert wrap('ab\x1b\x00cdef', 3) == ['ab\x1b\x00c', 'def']


def test_wrap_bare_esc_at_line_start():
    """
    Exercises the bare-ESC safety break in _find_first_visible_break.

    A bare \\x1b that does not match ZERO_WIDTH_PATTERN triggers a safety 'break' statement but is
    not found in any practical terminal sequence or string (ESC followed by NUL).
    """
    assert wrap('\x1b\x00あ', 1) == ['\x1b', '\x00', 'あ']



#!/usr/bin/env python
"""
Manual tests comparing wcwidth.py to libc's wcwidth(3) and wcswidth(3).

    https://github.com/jquast/wcwidth

This suite of tests compares the libc return values with the pure-python return
values. Although wcwidth(3) is POSIX, its actual implementation may differ,
so these tests are not guaranteed to be successful on all platforms, especially
where wcwidth(3)/wcswidth(3) is out of date. This is especially true for many
platforms -- usually conforming only to unicode specification 1.0 or 2.0.

This program accepts one optional command-line argument, the unicode version
level for our library to use when comparing to libc.
"""

# pylint: disable=C0103
#         Invalid module name "wcwidth-libc-comparator"

# standard imports

# std imports
import sys
import locale
import warnings
import ctypes.util
import unicodedata

# local
# local imports
import wcwidth


def is_named(ucs):
    """
    Whether the unicode point ``ucs`` has a name.

    :rtype bool
    """
    try:
        return bool(unicodedata.name(ucs))
    except ValueError:
        return False


def is_not_combining(ucs):
    return not unicodedata.combining(ucs)


def report_ucs_msg(ucs, wcwidth_libc, wcwidth_local):
    """
    Return string report of combining character differences.

    :param ucs: unicode point.
    :type ucs: unicode
    :param wcwidth_libc: libc-wcwidth's reported character length.
    :type comb_py: int
    :param wcwidth_local: wcwidth's reported character length.
    :type comb_wc: int
    :rtype: unicode
    """
    ucp = (ucs.encode('unicode_escape')[2:]
           .decode('ascii')
           .upper()
           .lstrip('0'))
    url = f"http://codepoints.net/U+{ucp}"
    name = unicodedata.name(ucs)
    return (
        f"libc,ours={wcwidth_libc},{wcwidth_local} "
        f"[--o{ucs}o--] name={name} val={ord(ucs)} {url} ")


if sys.maxunicode < 1114111:
    warnings.warn('narrow Python build, only a small subset of '
                  'characters may be tested.')


def _is_equal_wcwidth(libc, ucs, unicode_version):
    w_libc = libc.wcwidth(ucs)
    w_local = wcwidth.wcwidth(ucs, unicode_version)
    assert w_libc == w_local, report_ucs_msg(ucs, w_libc, w_local)


def main(using_locale=('en_US', 'UTF-8',)):
    """
    Program entry point.

    Load the entire Unicode table into memory, excluding those that:

        - are not named (func unicodedata.name returns empty string),
        - are combining characters.

    Using ``locale``, for each unicode character string compare libc's
    wcwidth with local wcwidth.wcwidth() function; when they differ,
    report a detailed AssertionError to stdout.
    """
    all_ucs = (ucs for ucs in
               [chr(val) for val in range(sys.maxunicode)]
               if is_named(ucs) and is_not_combining(ucs))

    libc_name = ctypes.util.find_library('c')
    if not libc_name:
        raise ImportError("Can't find C library.")

    libc = ctypes.cdll.LoadLibrary(libc_name)
    libc.wcwidth.argtypes = [ctypes.c_wchar, ]
    libc.wcwidth.restype = ctypes.c_int

    assert getattr(libc, 'wcwidth', None) is not None
    assert getattr(libc, 'wcswidth', None) is not None

    locale.setlocale(locale.LC_ALL, using_locale)
    unicode_version = 'latest'
    if len(sys.argv) > 1:
        unicode_version = sys.argv[1]

    for ucs in all_ucs:
        try:
            _is_equal_wcwidth(libc, ucs, unicode_version)
        except AssertionError as err:
            print(err)


if __name__ == '__main__':
    main()
